#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Base Classes '''

from __future__ import print_function, division, absolute_import, unicode_literals

import six
import weakref
from .utils.rest import RESTHelpers


def attribute(name, dtype=None, values=None):
    '''
    Create an XML attribute-based property

    Parameters
    ----------
    name : string
        The name of the XML attribute
    dtype : string, optional
        The data type of the attribute
        Valid values: int, float, double, string, bool
        Default: string
    values : list or dict, optional
        The valid values of the attribute.  If it is a list, the values
        in the list are used in both the instance attribute and the XML.
        If a dictionary is specified, the keys are the instance attribute
        value and the values are the XML values.

    Returns
    -------
    :class:`Attribute`

    '''
    return Attribute(name, dtype=dtype, values=values)


class Attribute(object):
    '''
    XML attribute-based property

    Parameters
    ----------
    name : string
        The name of the XML attribute
    dtype : string, optional
        The data type of the attribute
        Valid values: int, float, double, string, bool
        Default: string
    values : list or dict, optional
        The valid values of the attribute.  If it is a list, the values
        in the list are used in both the instance attribute and the XML.
        If a dictionary is specified, the keys are the instance attribute
        value and the values are the XML values.

    Returns
    -------
    :class:`Attribute`

    '''

    def __init__(self, name, dtype=None, values=None):
        self.name = name
        self.dtype = dtype or 'string'
        self.values = values
        self.value = weakref.WeakKeyDictionary()

    def get_xml_value(self, instance, owner=None):
        if instance is None:
            instance = self

        value = self.value.get(instance, None)
        if value is not None:

            if isinstance(self.values, dict):
                value = self.values[value]

            dtype = self.dtype
            if dtype == 'bool':
                value = value and 'true' or 'false'
            else:
                value = '%s' % value

        return value

    def get_value(self, instance, owner=None):
        if instance is None:
            instance = self
        return self.value.get(instance, None)

    def __get__(self, instance, owner=None):
        if instance is None:
            instance = self
        return self.value.get(instance, None)

    def __set__(self, instance, value):
        if instance is None:
            instance = self

        dtype = self.dtype

        if value is None:
            self.value[instance] = None
            return

        if dtype == 'int':
            value = int(value)

        elif dtype in ['double', 'float']:
            value = float(value)

        elif dtype == 'bool':
            if isinstance(value, six.string_types):
                value = (value.lower() == 'true')
            else:
                value = bool(value)

        elif dtype == 'string':
            value = '%s' % value

        if self.values:
            if isinstance(self.values, (tuple, list, set)):
                if value not in self.values:
                    raise ValueError('%s is not one of %s' % (value, self.values))
            elif isinstance(self.values, dict):
                if value not in self.values:
                    found = False
                    for key, val in self.values.items():
                        if value == val:
                            value = key
                            found = True
                            break
                    if not found:
                        raise ValueError('%s is not one of %s' %
                                         (value, list(self.values.keys())))

        self.value[instance] = value

    def __delete__(self, instance):
        if instance is None:
            instance = self
        self.value[instance] = None


class ESPObject(RESTHelpers):
    ''' Base class for all ESP objects '''

    def __init__(self, session=None, attrs=None):
        RESTHelpers.__init__(self, session=session)
        self._set_attributes(attrs)

    def _set_attributes(self, kwargs):
        kwargs = kwargs or {}
        xml_map = dict(getattr(type(self), 'xml_map', {}))

        # Always add these keys
        xml_map['name'] = 'name'
        xml_map['contquery'] = 'contquery'
        xml_map['project'] = 'project'

        attrs = {}

        for cls in reversed(type(self).__mro__):
            for key, value in vars(cls).items():
                if isinstance(value, Attribute):
                    attrs[key] = key
                    attrs[value.name] = key

        for key, value in kwargs.items():
            if value is not None and key in attrs:
                setattr(self, attrs[key], value)
            elif value is not None and key in xml_map:
                setattr(self, xml_map[key], value)

    def _get_attributes(self, use_xml_values=True):
        xml_map = dict(getattr(type(self), 'xml_map', {}))

        # Always add these keys
        xml_map['name'] = 'name'
        xml_map['contquery'] = 'contquery'
        xml_map['project'] = 'project'

        out = {}

        for cls in reversed(type(self).__mro__):
            for key, value in vars(cls).items():
                if isinstance(value, Attribute):
                    if use_xml_values:
                        val = value.get_xml_value(self)
                        if val is not None:
                            out[value.name] = val
                    else:
                        val = value.get_value(self)
                        if val is not None:
                            out[key] = val

        for attr_name, xml_name in xml_map.items():
            value = getattr(self, attr_name, None)
            if value is not None:
                if use_xml_values:
                    if type(value) is bool:
                        out[xml_name] = value and 'true' or 'false'
                    else:
                        out[xml_name] = '%s' % value
                else:
                    out[attr_name] = value

        return out

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return hash(other) == hash(self)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import copy
import datetime
import esppy
import os
import pandas as pd
import six
import sys
import time
import unittest
from esppy.config import ESP_ROOT
from esppy.plotting import ChartLayout
from esppy.windows import Subscriber, Publisher, Window
from esppy.windows.base import param_iter, var_mapper, Target
from esppy.utils import xml
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestParameterIterator(tm.TestCase):

    def test_empty(self):
        piter = param_iter({})
        self.assertEqual(next(piter), {})
        self.assertEqual(next(piter), {})

    def test_dict(self):
        for i, item in enumerate(param_iter(dict(a=[1, 2, 3], b='foo'))):
            if i == 0:
                self.assertEqual(item, dict(a=1, b='foo')) 
            elif i == 1:
                self.assertEqual(item, dict(a=2, b='foo')) 
            elif i == 2:
                self.assertEqual(item, dict(a=3, b='foo')) 
            elif i == 3:
                self.assertEqual(item, dict(a=1, b='foo')) 
                break

    def test_list(self):
        for i, item in enumerate(param_iter([dict(a=1, b='foo'), dict(b=2)])):
            if i == 0:
                self.assertEqual(item, dict(a=1, b='foo'))
            elif i == 1:
                self.assertEqual(item, dict(b=2))
            elif i == 2:
                self.assertEqual(item, dict(a=1, b='foo'))
                break


class TestUtils(tm.TestCase):

    def test_var_mapper(self):
        data = dict(num=[1, 2], char=['a', 'b'], bool=[True, False])

        df = pd.DataFrame.from_dict(data)

        # Dict mapping
        out = var_mapper(df,  dict(num=dict(outname={1:100, 2:200})))
        self.assertEqual(out, dict(outname=[100, 200]))

        # Function mapping
        def map_chars(x):
            out = []
            for item in x.values:
                if item == 'a':
                    out.append('FIRST')
                else:
                    out.append(item.upper())
            return out

        out = var_mapper(df,  dict(char=dict(outname=map_chars)))
        self.assertEqual(out, dict(outname=['FIRST', 'B']))


class TestTarget(tm.TestCase):

    def setUp(self):
        self.target1 = Target('T1', role='data', slot=0)
        self.target1a = Target('T1', role='data', slot=0)
        self.target2 = Target('T2', role='model', slot=1)
        self.target3 = Target('T3', role='left')
        self.target4 = Target('T4', role='right')
        self.target5 = Target('T5', slot=2)
        self.target6 = Target('T6')

    def test_equality(self):
        self.assertEqual(self.target1, self.target1) 
        self.assertNotEqual(self.target1, self.target1a) 

    def test_str(self):
        self.assertEqual(str(self.target1), "Target('T1', role='data', slot=0)")
        self.assertEqual(str(self.target2), "Target('T2', role='model', slot=1)")
        self.assertEqual(str(self.target3), "Target('T3', role='left')")
        self.assertEqual(str(self.target4), "Target('T4', role='right')")
        self.assertEqual(str(self.target5), "Target('T5', slot=2)")
        self.assertEqual(str(self.target6), "Target('T6')")

    def test_repr(self):
        self.assertEqual(repr(self.target1), "Target('T1', role='data', slot=0)")
        self.assertEqual(repr(self.target2), "Target('T2', role='model', slot=1)")
        self.assertEqual(repr(self.target3), "Target('T3', role='left')")
        self.assertEqual(repr(self.target4), "Target('T4', role='right')")
        self.assertEqual(repr(self.target5), "Target('T5', slot=2)")
        self.assertEqual(repr(self.target6), "Target('T6')")


class TestSubscriber(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']
        self.window = self.project.queries['contquery']['w_data']
        self.sub = None

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")
        if self.sub is not None:
            self.sub.close()

    def test_init(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        events = dict(message=None, event=None, open=None, close=None)

        def on_event(sock, event):
            events['event'] = event

        def on_message(sock, message):
            events['message'] = message

        def on_open(sock):
            events['open'] = True

        def on_close(sock, code, reason=None):
            events['close'] = True

        self.sub = Subscriber(self.window,
                              mode='streaming',
                              on_event=on_event,
                              on_message=on_message,
                              on_close=on_close,
                              on_open=on_open)
        self.sub.start()
        self.window.publish_events(data_path)
        time.sleep(3) 
        self.sub.close()

        # Wait for on_close event
        count = 0
        while events['close'] is not True:
            time.sleep(2) 
            count += 1
            if count >= 15:
                break

        self.assertTrue(isinstance(events['event'], pd.DataFrame))
        self.assertTrue(isinstance(events['message'], six.string_types))
        self.assertTrue(events['open'] is True)
        self.assertTrue(events['close'] is True)

    def test_url(self):
        sub = Subscriber(self.window)
        # NOTE: schema is always forced to true internally
        self.assertEqual(sub.url.split('/%s/subscribers/' % ESP_ROOT, 1)[-1],
                         'ESPUnitTestProjectSA/contquery/w_data/?'
                         'format=xml&mode=updating&pagesize=50&schema=true')

    def test_properties(self):
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        self.sub = sub = Subscriber(self.window)

        self.assertFalse(sub.is_active)

        sub.start()
        self.window.publish_events(data_path)
        time.sleep(3) 

        self.assertTrue(sub.is_active)

        sub.mode = 'streaming'
        sub.pagesize = 10
        sub.sort = 'id'
        sub.interval = 200
        sub.filter = 'in(id, 1, 2, 3, 4, 5, 6)'
        sub.separator = ';'

        self.assertEqual(sub.mode, 'streaming')
        self.assertEqual(sub.pagesize, 10)
        self.assertEqual(sub.sort, 'id')
        self.assertEqual(sub.interval, 200)
        self.assertEqual(sub.filter, 'in(id, 1, 2, 3, 4, 5, 6)')
        self.assertEqual(sub.separator, ';')

        time.sleep(3) 


class TestPublisher(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")


class TestWindow(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        try:
            self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")
        except:
            pass

    def test_schema(self):
        win = Window(schema='foo*:int64,bar:int32,baz:double') 
        self.assertEqual(win.schema.schema_string, 'foo*:int64,bar:int32,baz:double')

        win.schema = 'a:double', 'b:double'
        self.assertEqual(win.schema.schema_string, 'a:double,b:double')

    def test_subscriber_url(self):
        self.assertEqual(self.query['w_data'].subscriber_url.split('/%s/subscribers/' % ESP_ROOT)[-1],
                         'ESPUnitTestProjectSA/contquery/w_data/')

    def test_publisher_url(self):
        self.assertEqual(self.query['w_data'].publisher_url.split('/%s/publishers/' % ESP_ROOT)[-1],
                         'ESPUnitTestProjectSA/contquery/w_data/')

    def test_event_horizon(self):
        win = self.query['w_data']

        # TODO: Need a way to verify

        # Datetime
        win.subscribe(horizon=datetime.datetime.now() + datetime.timedelta(hours=1))
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

        # Time delta
        win.subscribe(horizon=datetime.timedelta(hours=1))
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

        # Date
        win.subscribe(horizon=(datetime.datetime.now() + datetime.timedelta(hours=1)).date())
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

        # Time
        win.subscribe(horizon=(datetime.datetime.now() + datetime.timedelta(hours=1)).time())
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

        # Number of events
        win.subscribe(horizon=20)
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

        # Unknown type
        win.subscribe(win)
        time.sleep(2)
        win.unsubscribe()
        time.sleep(1)

    def test_add_event_transformer(self):
        win = self.query['w_data']

        def square_x(event):
            event['x_c'] = event['x_c'] * event['x_c']
            return event

        win.add_event_transformer('clip', upper=4)
        win.add_event_transformer(square_x)

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')

        # Start subscriber
        win.subscribe()
        time.sleep(2)
        
        win.publish_events(data_path)
        time.sleep(2)

        self.assertEqual(list(win.data.index), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        self.assertEqual(win['x_c'].tolist(), [1, 4, 9, 16, 16, 16, 9, 16, 4, 16, 1])
        self.assertEqual(win['y_c'].tolist(), [1, 1, 0, 4, 4, 4, 1, 4, 2, 4, 3])

        def negate_x(event):
            event['x_c'] = -event['x_c']
            return event

        win.add_event_transformer(negate_x)

        self.assertEqual(list(win.data.index), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        self.assertEqual(win['x_c'].tolist(), [-1, -4, -9, -16, -16, -16, -9, -16, -4, -16, -1])
        self.assertEqual(win['y_c'].tolist(), [1, 1, 0, 4, 4, 4, 1, 4, 2, 4, 3])

        win.add_event_transformer('abs')

        self.assertEqual(list(win.data.index), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        self.assertEqual(win['x_c'].tolist(), [1, 4, 9, 16, 16, 16, 9, 16, 4, 16, 1])
        self.assertEqual(win['y_c'].tolist(), [1, 1, 0, 4, 4, 4, 1, 4, 2, 4, 3])

        # TODO: Remove once server crash is fixed
        win._subscriber.close()

    def test_create_subscriber(self):
        win = self.query['w_data']

        out = dict(message=None, event=None, open=None, close=None, error=None)

        def on_message(sock, message):
            out['message'] = message

        def on_event(sock, event):
            out['event'] = event

        def on_open(sock):
            out['open'] = True

        def on_close(sock, code, reason=None):
            out['close'] = True

        def on_error(sock, error):
            out['error'] = error 

        sub = win.create_subscriber(on_message=on_message,
                                    on_event=on_event,
                                    on_open=on_open,
                                    on_close=on_close,
                                    on_error=on_error)
        sub.start()
        time.sleep(2)

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(data_path)
        time.sleep(2)

        sub.close()

        # Wait for on_close event
        count = 0
        while out['close'] is not True:
            time.sleep(2) 
            count += 1
            if count >= 15:
                break

        self.assertTrue(isinstance(out['message'], six.string_types))
        self.assertTrue(isinstance(out['event'], pd.DataFrame))
        self.assertTrue(out['open'] is True)
        self.assertTrue(out['close'] is True)
        self.assertTrue(out['error'] is None)
        
    def test_create_publisher(self):
        win = self.query['w_data']

        out = dict(events=None)

        def on_event(sock, event):
            out['events'] = event

        sub = win.create_subscriber(on_event=on_event)
        sub.start()
        pub = win.create_publisher(format='csv', blocksize=10)

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        pub.send(open(data_path).read())
        time.sleep(5)
        pub.close()

        self.assertTrue(isinstance(out['events'], pd.DataFrame))
        self.assertEqual(len(out['events']), 10)

        sub.close()

    def test_publish_events(self):
        win = self.query['w_data']

        out = dict(msg=None)

        def on_message(sock, message): 
            out['msg'] = message 

        sub = win.create_subscriber(on_message=on_message, format='csv')
        sub.start()

        # File path
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(data_path)
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # File object
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(open(data_path))
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # CSV
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(open(data_path).read())
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # DataFrame
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        df = pd.read_csv(data_path, header=None,
                                    names=['opcode', 'opmode', 'id', 'x_c', 'y_c'])
        win.publish_events(df)
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        del df['opcode']
        del df['opmode']

        # XML
        xml = []
        for row in df.itertuples(index=False):
            xml.append('<event opcode="insert">')
            for col, value in zip(df.columns, row):
                xml.append('  <%s>%s</%s>' % (col, value, col))
            xml.append('</event>')
            xml.append('')
        xml = '\n'.join(xml)
        win.publish_events(xml)
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # JSON
        json = []
        for row in df.itertuples(index=False):
            json.append('{')
            json.append('  "opcode":"insert",')
            json.append('  "event":{')
            for col, value in zip(df.columns, row):
                json.append('  "%s":"%s",' % (col, value))
            json[-1] = json[-1][:-1]
            json.append('  }')
            json.append('}')
            json.append('')
        json = '\n'.join(json)
        win.publish_events(json)
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # Properties
        prop = []
        for row in df.itertuples(index=False):
            prop.append('opcode=insert')
            for col, value in zip(df.columns, row):
                prop.append('%s=%s' % (col, value))
            prop.append('')
        prop.append('')
        prop = '\n'.join(prop)
        win.publish_events(prop)
        time.sleep(3)
        self.assertEqual(out['msg'].strip(), 'I,N, 11,1.000000,3.000000')
        out['msg'] = None

        # Unknown
        with self.assertRaises(TypeError):
            win.publish_events(None)

        sub.close()

    def test_apply_transformers(self):
        win = self.query['w_data']

        def neg_square_x(event):
            event['x_c'] = -event['x_c'] * event['x_c']
            return event

        win.add_event_transformer(neg_square_x)

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        df = pd.read_csv(data_path, header=None, names=['opcode', 'opmode', 'id', 'x_c', 'y_c'])
        df = df.set_index('id')
        del df['opcode']
        del df['opmode']

        out = win.apply_transformers(df)
        self.assertEqual(out.to_csv(),
                         'id,x_c,y_c\n'
                         '1,-1,1\n'
                         '2,-4,1\n'
                         '3,-9,0\n'
                         '4,-36,5\n'
                         '5,-36,6\n'
                         '6,-49,5\n'
                         '7,-9,1\n'
                         '8,-25,5\n'
                         '9,-4,2\n'
                         '10,-49,4\n'
                         '11,-1,3\n')

        win.event_transformers[:] = []
        win.add_event_transformer('abs')

        out = win.apply_transformers(df)
        self.assertEqual(out.to_csv(),
                         'id,x_c,y_c\n'
                         '1,1,1\n'
                         '2,4,1\n'
                         '3,9,0\n'
                         '4,36,5\n'
                         '5,36,6\n'
                         '6,49,5\n'
                         '7,9,1\n'
                         '8,25,5\n'
                         '9,4,2\n'
                         '10,49,4\n'
                         '11,1,3\n')

    def test_contquery(self):
        win = self.query['w_data']
        self.assertEqual(win.contquery, 'contquery')
        self.assertEqual(win.project, 'ESPUnitTestProjectSA')

        win.contquery = 'foo'
        win.project = 'bar'
        self.assertEqual(win.contquery, 'foo')
        self.assertEqual(win.project, 'bar')

        win.contquery = self.query
        self.assertEqual(win.contquery, 'contquery')
        self.assertEqual(win.project, 'ESPUnitTestProjectSA')

    def test_project(self):
        win = self.query['w_data']
        self.assertEqual(win.project, 'ESPUnitTestProjectSA')

        win.project = 'foo'
        self.assertEqual(win.project, 'foo')

        win.project = self.project
        self.assertEqual(win.project, 'ESPUnitTestProjectSA')

    @unittest.skip('Event generators don\'t work with embedded data')
    def test_create_event_generator(self):
        win = self.query['w_data']
        win.subscribe()

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        egen = win.create_event_generator(data_path)
        egen.start()
        egen.delete()

        time.sleep(20)
        print(win)
        self.assertEqual(len(win), 11)

        win.unsubscribe()

    def test_add_targets(self):
        win = self.query['w_data']
        win.add_targets('foo', 'bar', role='data')
        targets = [(x.name, x.role) for x in win.targets]
        self.assertIn(('w_calculate', 'data'), targets)
        self.assertIn(('foo', 'data'), targets)
        self.assertIn(('bar', 'data'), targets)

    def test_delete_targets(self):
        win = self.query['w_data']
        win.add_targets('foo', 'bar', role='data')
        targets = [(x.name, x.role) for x in win.targets]
        self.assertIn(('w_calculate', 'data'), targets)
        self.assertIn(('foo', 'data'), targets)
        self.assertIn(('bar', 'data'), targets)

        win.delete_targets('foo', 'w_calculate')
        targets = [(x.name, x.role) for x in win.targets]
        self.assertNotIn(('w_calculate', 'data'), targets)
        self.assertNotIn(('foo', 'data'), targets)
        self.assertIn(('bar', 'data'), targets)

    def test_fullname(self):
        self.assertEqual(self.query['w_data'].fullname,
                         'ESPUnitTestProjectSA.contquery.w_data')

    def test_schema_string(self):
        win = self.query['w_data']
        self.assertEqual(win.schema_string, 'id*:int64,x_c:double,y_c:double')

        win.schema.fields.clear()
        self.assertEqual(win.schema_string, '')

        win.schema_string = 'foo:double,bar:int64'
        self.assertEqual(len(win.schema.fields), 2)
        self.assertEqual(win.schema.fields['foo'].type, 'double')
        self.assertEqual(win.schema.fields['bar'].type, 'int64')

    def test_url(self):
        self.assertEqual(self.query['w_data'].url.split('/%s/' % ESP_ROOT, 1)[-1],
                         'ESPUnitTestProjectSA/contquery/w_data/') 

    def test__verify_project(self):
        win = self.query['w_data']
        self.assertTrue(win._verify_project() is None)

        win.project = None
        with self.assertRaises(ValueError):
            win._verify_project()

        win.project = self.project
        win.contquery = None
        with self.assertRaises(ValueError):
            win._verify_project()

    def test_copy(self):
        win = self.query['w_data']

        # Using copy method
        wcopy = win.copy()
        self.assertEqual(wcopy.project, win.project)
        self.assertEqual(wcopy.contquery, win.contquery)
        self.assertEqual(wcopy.schema_string, win.schema_string)
        self.assertEqual(wcopy.event_transformers, win.event_transformers)

        # Using copy method with deep=True
        wcopy = win.copy(deep=True)
        self.assertEqual(wcopy.project, win.project)
        self.assertEqual(wcopy.contquery, win.contquery)
        self.assertEqual(wcopy.schema_string, win.schema_string)
        self.assertTrue(wcopy.schema is not win.schema)
        self.assertEqual(wcopy.event_transformers, win.event_transformers)

        # Using copy.copy
        wcopy = copy.copy(win)
        self.assertEqual(wcopy.project, win.project)
        self.assertEqual(wcopy.contquery, win.contquery)
        self.assertEqual(wcopy.schema_string, win.schema_string)
        self.assertEqual(wcopy.event_transformers, win.event_transformers)

        # Using copy.deepcopy
        wcopy = copy.deepcopy(win)
        self.assertEqual(wcopy.project, win.project)
        self.assertEqual(wcopy.contquery, win.contquery)
        self.assertEqual(wcopy.schema_string, win.schema_string)
        self.assertTrue(wcopy.schema is not win.schema)
        self.assertEqual(wcopy.event_transformers, win.event_transformers)

    def test_to_element(self):
        win = self.query['w_data']
        elem = win.to_element()
        out = xml.to_xml(elem, pretty=True)
        self.assertEqual(out.strip(), '''<window-source index="pi_EMPTY" insert-only="true" name="w_data">
  <schema>
    <fields>
      <field key="true" name="id" type="int64" />
      <field key="false" name="x_c" type="double" />
      <field key="false" name="y_c" type="double" />
    </fields>
  </schema>
  <connectors>
    <connector class="fs" name="publisher" type="publish">
      <properties>
        <property name="blocksize">1</property>
        <property name="fsname">/u/kesmit/gitlab/python-esp/esppy/tests/data//data_sa.csv</property>
        <property name="fstype">csv</property>
        <property name="transactional">true</property>
      </properties>
    </connector>
  </connectors>
</window-source>''')

    def test_to_xml(self):
        win = self.query['w_data']
        out = win.to_xml(pretty=True)
        self.assertEqual(out.strip(), '''<window-source index="pi_EMPTY" insert-only="true" name="w_data">
  <schema>
    <fields>
      <field key="true" name="id" type="int64" />
      <field key="false" name="x_c" type="double" />
      <field key="false" name="y_c" type="double" />
    </fields>
  </schema>
  <connectors>
    <connector class="fs" name="publisher" type="publish">
      <properties>
        <property name="blocksize">1</property>
        <property name="fsname">/u/kesmit/gitlab/python-esp/esppy/tests/data//data_sa.csv</property>
        <property name="fstype">csv</property>
        <property name="transactional">true</property>
      </properties>
    </connector>
  </connectors>
</window-source>''')

    def test_to_graph(self):
        # TODO: Not sure how to verify this
        grph = self.query['w_data'].to_graph(schema=True)

    def test_repr_svg(self):
        # TODO: Not sure how to verify this
        grph = self.query['w_data']._repr_svg_()

    def test_mapping_methods(self):
        win = self.query['w_data']
        win.subscribe()
 
        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(data_path)
        time.sleep(3)

        self.assertTrue(win['x_c'] is not None)
        time.sleep(3)
        self.assertEqual(len(win['x_c']), 11)

        with self.assertRaises(RuntimeError):
            win['x_c'] = range(100, 111)

        with self.assertRaises(RuntimeError):
             del win['x_c']

        self.assertTrue(isinstance(win.x_c, pd.Series))
        self.assertTrue(isinstance(win.data, pd.DataFrame))

        self.assertEqual(list(iter(win)), ['x_c', 'y_c'])

    def test_enable_tracing(self):
        self.query['w_data'].enable_tracing()

    def test_disable_tracing(self):
        self.query['w_data'].disable_tracing()

    @unittest.skip('Hangs server')
    def test_get_events(self):
        win = self.query['w_data']

        data_path = os.path.join(DATA_DIR, 'data_sa.csv')
        win.publish_events(data_path)
        time.sleep(5)

        out = win.get_events()
        self.assertTrue(isinstance(out, pd.DataFrame))
        self.assertEqual(list(out.columns), ['x_c', 'y_c'])
        self.assertEqual(list(out.index.names), ['id'])
        self.assertEqual(len(out), 11)

    def test_str(self):
        self.assertEqual(str(self.query['w_data']),
                         "SourceWindow(name='w_data', contquery='contquery',"
                         " project='ESPUnitTestProjectSA')")

    def test_repr(self):
        self.assertEqual(repr(self.query['w_data']),
                         "SourceWindow(name='w_data', contquery='contquery',"
                         " project='ESPUnitTestProjectSA')")

    def test_to_data_callback(self):
        win = self.query['w_data']

        def var_generator(data):
            return dict(sum=data['x_c'] + data['y_c'])

        out = win.to_data_callback(x='id', y=['x_c', 'y_c'], var_generator=var_generator)
        data = out(initial=True)
        self.assertEqual(sorted(data.keys()), ['id', 'sum', 'x_c', 'y_c'])
        data = out()
        self.assertEqual(sorted(data.keys()), ['id', 'sum', 'x_c', 'y_c'])
        data = out(terminate=True)
        self.assertTrue(data is None)

        out = win.to_data_callback(x=['id'], y='x_c', extra='y_c')
        data = out(initial=True)
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out()
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out(terminate=True)
        self.assertTrue(data is None)

        out = win.to_data_callback(x=['id'], y='x_c', extra=['y_c'])
        data = out(initial=True)
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out()
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out(terminate=True)
        self.assertTrue(data is None)

        out = win.to_data_callback(extra=['id', 'x_c', 'y_c'])
        data = out(initial=True)
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out()
        self.assertEqual(sorted(data.keys()), ['id', 'x_c', 'y_c'])
        data = out(terminate=True)
        self.assertTrue(data is None)

    def test_streaming_line(self):
        fig = self.query['w_data'].streaming_line(x='id', y=['x_c', 'y_c'])
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_line(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_scatter(self):
        fig = self.query['w_data'].streaming_scatter(x='id', y=['x_c', 'y_c'])
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_scatter(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_area(self):
        fig = self.query['w_data'].streaming_area(x='id', y=['x_c', 'y_c'])
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_area(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_bubble(self):
        fig = self.query['w_data'].streaming_bubble(x='id', y=['x_c', 'y_c'], radius='id')
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_bubble(x='id', y='x_c', radius='id')
        out = fig._repr_html_()

    def test_streaming_bar(self):
        fig = self.query['w_data'].streaming_bar(x='id', y=['x_c', 'y_c'])
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_bar(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_hbar(self):
        fig = self.query['w_data'].streaming_hbar(x='id', y=['x_c', 'y_c'])
        out = fig._repr_html_()

        fig = self.query['w_data'].streaming_hbar(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_donut(self):
        fig = self.query['w_data'].streaming_donut(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_pie(self):
        fig = self.query['w_data'].streaming_pie(x='id', y='x_c')
        out = fig._repr_html_()

    def test_streaming_images(self):
        fig = self.query['w_data'].streaming_images('x_c')
        out = fig._repr_html_()

    def test_chart_layout(self):
        fig1 = self.query['w_data'].streaming_line(x='id', y=['x_c', 'y_c'])
        fig2 = self.query['w_data'].streaming_line(x='id', y=['x_c', 'y_c'])
        fig3 = self.query['w_data'].streaming_line(x='id', y=['x_c', 'y_c'])

        layout = ChartLayout(fig1, (fig2, fig3)) 
        
        out = layout._repr_html_()

    def test_streaming_hist(self):
        fig = self.query['w_data'].streaming_hist(centers=['c1', 'c2', 'c3'],
                                                  heights=['h1', 'h2', 'h3'])
        out = fig._repr_html_()

    def test_copyvars(self):
        # Specified columns
        win = self.s.CalculateWindow(input_map=dict(a='A1', b='B1', c='C1'),
                                     output_map=dict(x='X1', y='Y1', z='Z1'))

        win.copyvars = ['A1', 'C1']
        self.assertEqual(list(win.schema.fields.keys()), ['A1', 'C1'])

        # With key
        win = self.s.CalculateWindow(input_map=dict(a='A1*', b='B1', c='C1'),
                                     output_map=dict(x='X1', y='Y1', z='Z1'))

        win.copyvars = ['A1', 'C1']
        self.assertEqual(list(win.schema.fields.keys()), ['A1', 'C1'])
        self.assertTrue(win.schema['A1'].key)
        self.assertFalse(win.schema['C1'].key)

        # With data type
        win = self.s.CalculateWindow(input_map=dict(a='A1:int32', b='B1', c='C1'),
                                     output_map=dict(x='X1', y='Y1', z='Z1'))

        win.copyvars = ['A1', 'C1']
        self.assertEqual(list(win.schema.fields.keys()), ['A1', 'C1'])
        self.assertEqual(win.schema['A1'].type, 'int32')
        self.assertEqual(win.schema['C1'].type, 'double')

        # With data type and key
        win = self.s.CalculateWindow(input_map=dict(a='A1*:int32', b='B1', c='C1'),
                                     output_map=dict(x='X1', y='Y1', z='Z1'))

        win.copyvars = ['A1', 'C1']
        self.assertEqual(list(win.schema.fields.keys()), ['A1', 'C1'])
        self.assertEqual(win.schema['A1'].type, 'int32')
        self.assertEqual(win.schema['C1'].type, 'double')
        self.assertTrue(win.schema['A1'].key)
        self.assertFalse(win.schema['C1'].key)

        # With array type
        win = self.s.CalculateWindow(input_map=dict(a='A1[1-5]:int32', b='B1', c='C1'),
                                     output_map=dict(x='X1', y='Y1', z='Z1'))

        win.copyvars = ['A1', 'C1']
        self.assertEqual(list(win.schema.fields.keys()), ['A1', 'C1'])
        self.assertEqual(win.schema['A1'].type, 'array(i32)')
        self.assertEqual(win.schema['C1'].type, 'double')


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import os
import six
import esppy
import numpy as np
import pandas as pd
import sys
import unittest
from PIL import Image
from esppy.plotting import split_chart_params, highlight_image
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestPlottingUtils(tm.TestCase):

    def test_split_chart_params(self):
        fig, mthd = split_chart_params(x_range=[0,1], y_range=[0,1], x='x_c', y='y_c')
        self.assertEqual(fig, dict(x_range=[0,1], y_range=[0,1]))
        self.assertEqual(mthd, dict(x='x_c', y='y_c'))

    def test_highlight_image(self):
        in_img = Image.open(os.path.join(DATA_DIR, 'scooter.jpg'))

        detections = pd.DataFrame([[1, 'Scooter', 1, 35, 35, 100, 100]],
                                  columns=['_nObjects_', '_Object0_', '_P_Object0_',
                                           '_Object0_x', '_Object0_y', '_Object0_width',
                                           '_Object0_height'])

        out_img = highlight_image(in_img, detections)

        bench_img = Image.open(os.path.join(DATA_DIR, 'scooter-hilite.jpg'))

#       TODO: Compare with benchmark
#       self.assertEqual(list(out_img.getdata()), list(bench_img.getdata()))


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Window Subscriber '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import copy
import csv
import datetime
import functools
import itertools
import os
import pandas as pd
import re
import requests
import six
import sys
import threading
import types
import weakref
import xml.etree.ElementTree as ET
from ..utils.authorization import Authorization
from six.moves import urllib
from .utils import verify_window
from ..base import ESPObject, attribute
from ..config import get_option, CONCAT_OPTIONS
from ..exceptions import ESPError
from ..schema import Schema
from ..utils.keyword import dekeywordify
from ..utils import xml
from ..utils.notebook import scale_svg
from ..utils.rest import get_params
from ..utils.data import get_project_data, gen_name, get_server_info
from ..utils.events import get_events, get_dataframe, get_schema
from ..websocket import WebSocketClient


class Subscriber(object):
    '''
    Create a subscriber for the given window

    Attributes
    ----------
    callbacks : dict
        The dictionary of callback functions
    filter : string
        Functional filter to subset events
    format : string, optional
        The format of the received data: 'xml', 'json', 'csv', 'properties'
    interval : int
        Interval between event sends in milliseconds
    is_active : bool
        Is the web socket currently active?
    mode : string
        The mode of subscriber: 'updating' or 'streaming'
    pagesize : int
        The maximum number of events in a page
    separator : string, optional
        The separator to use between events in the 'properties' format
    schema : bool
        Should the schema be sent with the first event?
    sort : string
        Sort order for the events (updating mode only)
    window_schema : Schema
        The schema of the window being subscribed to
    window_url : string
        The subscriber URL of the window

    Parameters
    ----------
    window : Window
        The window object to create the subscriber for
    mode : string, optional
        The mode of subscriber: 'updating' or 'streaming'
    pagesize : int, optional
        The maximum number of events in a page
    filter : string, optional
        Functional filter to subset events
    sort : string, optional
        Sort order for the events (updating mode only)
    format : string, optional
        The format of the received data: 'xml', 'json', 'csv', 'properties'
    separator : string, optional
        The separator to use between events in the 'properties' format
    interval : int, optional
        Interval between event sends in milliseconds
    schema : bool, optional
        Should the schema be sent with the first event?
    on_event : callable, optional
        The object to call for events. The argument to this object will
        be a DataFrame of the events that occurred.
    on_message : callable, optional
        The object to call for each websocket message
    on_error : callable, optional
        The object to call for each websocket error
    on_close : callable, optional
        The object to call when the websocket is opened
    on_open : callable, optional
        The object to call when the websocket is closed

    Examples
    --------
    Create event callback; ``event`` is a DataFrame

    >>> def on_event(event):
    ...     print(event.columns)

    Create the subscriber instance

    >>> sub = Subscriber(window, on_event=on_event)

    Start event processing (runs a thread in the background)

    >>> sub.start()

    Stop event processing (stops background thread)

    >>> sub.stop()

    Returns
    -------
    :class:`Subscriber`

    '''

    def __init__(self, window, mode='updating', pagesize=50, filter=None,
                 sort=None, format='xml', separator=None, interval=None,
                 schema=False, on_event=None, on_message=None, on_error=None,
                 on_close=None, on_open=None):
        self._ws = None
        self.mode = mode
        self.pagesize = pagesize
        self.filter = filter
        self.sort = sort
        self.format = format
        self.separator = separator
        self.interval = interval
        self.schema = schema
        self.server_info = get_server_info(window)
        self.session = window.session
        self.window_schema = get_schema(window, window)
        self.window_url = window.subscriber_url
        self.window_fullname = window.fullname
        self.callbacks = {k: v for k, v in dict(on_message=on_message,
                                                on_error=on_error,
                                                on_event=on_event,
                                                on_close=on_close,
                                                on_open=on_open).items()
                          if v is not None}

    @property
    def url(self):
        '''
        Return the URL of the subscriber

        Returns
        -------
        string

        '''
        url_params = get_params(**{'mode': self.mode,
                                   'pagesize': self.pagesize,
                                   'filter': self.filter,
                                   'format': self.format,
                                   'separator': self.separator,
                                   'sort': self.sort,
                                   'interval': self.interval,
                                   'schema': True})
        url_params = '&'.join(['%s=%s' % (k, v) for k, v in sorted(url_params.items())])
        return self.window_url + '?' + url_params

    @property
    def is_active(self):
        '''
        Is the web socket active?

        Returns
        -------
        bool

        '''
        return self._ws is not None

    @property
    def mode(self):
        '''
        The mode of the subscriber: 'updating' or 'streaming'

        Returns
        -------
        string

        '''
        return self._mode

    @mode.setter
    def mode(self, value):
        ''' Set the mode of the subscriber '''
        self._mode = value
        if self._ws is not None and self._mode is not None:
            self._ws.send('<properties mode="%s"></properties>' % self._mode)

    @property
    def pagesize(self):
        '''
        The maximum number of events in a page

        Returns
        -------
        int

        '''
        return self._pagesize

    @pagesize.setter
    def pagesize(self, value):
        ''' Set the pagesize '''
        self._pagesize = value
        if self._ws is not None and self._pagesize is not None:
            self._ws.send('<properties pagesize="%s"></properties>' % self._pagesize)

    @property
    def sort(self):
        '''
        Sort order for the events (updating mode only)

        Returns
        -------
        string

        '''
        return self._sort

    @sort.setter
    def sort(self, value):
        ''' Set the sort order for the events '''
        self._sort = value
        if self._ws is not None and self._sort is not None:
            self._ws.send('<properties sort="%s"></properties>' % self._sort)

    @property
    def interval(self):
        '''
        Interval between event sends in milliseconds

        Returns
        -------
        int

        '''
        return self._interval

    @interval.setter
    def interval(self, value):
        ''' Set the event interval '''
        self._interval = value
        if self._ws is not None and self._interval is not None:
            self._ws.send('<properties interval="%s"></properties>' % self._interval)

    @property
    def filter(self):
        '''
        Functional filter to subset events

        Returns
        -------
        string

        '''
        return self._filter

    @filter.setter
    def filter(self, value):
        ''' Set the filter string '''
        self._filter = value
        if self._ws is not None and self._filter is not None:
            self._ws.send(('<properties><filter><![CDATA[%s]]>'
                           '</filter></properties>') % self._filter)

    @property
    def separator(self):
        '''
        Separator to use between events in the 'properties' format

        Returns
        -------
        string

        '''
        return self._separator

    @separator.setter
    def separator(self, value):
        ''' Set the separator string '''
        self._separator = value
        if self._ws is not None and self._separator is not None:
            self._ws.send('<properties separator="%s"></properties>' % self._separator)

    def start(self):
        '''
        Initialize the web socket and start it in its own thread

        Notes
        -----
        The thread created in the background will continue to run unless
        explicitly stopped using the :meth:`stop` method.

        Examples
        --------
        Create subscriber instance

        >>> sub = Subscriber(window, on_event=on_event)

        Start processing events

        >>> sub.start()

        '''
        if self._ws is not None:
            return

        if not verify_window(self.window_fullname, session=self.session):
            raise ESPError('There is no window at %s' % self.window_fullname)

        state = dict(status=None, schema=None, dataframe=None)

        def on_message(sock, message):
            # HTTP status messages
            if state['status'] is None and re.match(r'^\s*\w+\s*\:\s*\d+\s*\n', message):
                state['status'] = int(re.match(r'^\s*\w+\s*\:\s*(\d+)', message).group(1))
                if state['status'] >= 400:
                    raise ESPError('Subscriber message returned with status: %s' %
                                   state['status'])
                return

            if state['schema'] is None:
                if message.startswith('<schema>'):
                    state['schema'] = Schema.from_xml(message)

                elif re.match(r'^\s*{\s*["\']?schema["\']?\s*:', message):
                    state['schema'] = Schema.from_json(message)

                else:
                    raise ValueError('Unrecognized schema definition format: %s...' %
                                     message[:40])

                state['dataframe'] = get_dataframe(state['schema'])
                if self.schema:
                    return message
                return

            if 'on_message' in self.callbacks:
                self.callbacks['on_message'](sock, message)

            if 'on_event' in self.callbacks:
                try:
                    df = get_events(state['schema'], message,
                                    single=True, format=self.format,
                                    separator=self.separator,
                                    server_info=self.server_info)
                    #self.callbacks['on_event'](sock, pd.concat([state['dataframe'], df], **CONCAT_OPTIONS))
                    self.callbacks['on_event'](sock, pd.concat([state['dataframe'], df]))
                except:
                    import traceback
                    traceback.print_exc()
                    raise

        def on_error(sock, error):
            if 'on_error' in self.callbacks:
                self.callbacks['on_error'](sock, error)

        def on_open(sock):
            if 'on_open' in self.callbacks:
                self.callbacks['on_open'](sock)

        def on_close(sock, code, reason=None):
            if 'on_close' in self.callbacks:
                self.callbacks['on_close'](sock, code, reason=None)

        if get_option('debug.requests'):
            sys.stderr.write('WEBSOCKET %s\n' % self.url)

        headers = []

        auth = Authorization.getInstance(self.session)

        if auth.isEnabled:
            headers.append(("Authorization",auth.authorization));

        self._ws = WebSocketClient(self.url,
                                   on_message=on_message,
                                   on_error=on_error,
                                   on_open=on_open,
                                   on_close=on_close,
                                   headers=headers)

        self._ws.connect()

        ws_thread = threading.Thread(name='%s-%s' % (id(self), self.window_url),
                                     target=self._ws.run_forever)
        ws_thread.daemon = True
        ws_thread.start()

    def stop(self):
        '''
        Stop processing events and close the web socket

        Examples
        --------
        Create subscriber instance

        >>> sub = Subscriber(window, on_event=on_event)

        Start processing events

        >>> sub.start()

        Stop processing events

        >>> sub.stop()

        '''
        if self._ws is not None:
            self._ws.close()
            self._ws = None

    close = stop



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import BaseWindow, attribute
from .features import ParametersFeature, ConnectorsFeature
from .utils import get_args


class ModelReaderWindow(BaseWindow, ParametersFeature, ConnectorsFeature):
    '''
    Model reader window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
    description : string, optional
        Description of the window
    model_type : string, optional
        Type of the model.
        Valid Values: astore or recommender

    Returns
    -------
    :class:`ModelReaderWindow`

    '''

    window_type = 'model-reader'

    model_type = attribute('model-type', dtype='string',
                           values=['astore', 'recommender'])

    def __init__(self, name=None, pubsub=None, description=None, model_type=None):
        BaseWindow.__init__(self, **get_args(locals()))



from urllib.parse import urlparse

import esppy.espapi.connections as connections

def connect(url,delegate = None,**kwargs):

    u = urlparse(url)

    secure = False

    if u[0] == "https":
        secure = True

    s = u[1].split(":")

    host = s[0]
    port = s[1]

    conn = connections.ServerConnection(host,port,secure,delegate,**kwargs)

    conn.start()

    return(conn)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Template '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
import warnings
from six.moves import urllib
from ..base import ESPObject, attribute
from ..config import get_option
from ..mas import MASModule
from ..windows import Target, BaseWindow, get_window_class, CalculateWindow, TrainWindow, ScoreWindow, ModelReaderWindow
from ..utils import xml
from ..utils.data import gen_name
from ..utils.notebook import scale_svg


class WindowDict(collections.MutableMapping):
    '''
    Dictionary for holding window objects

    Attributes
    ----------
    project : string
        The name of the project
    contquery : string
        The name of the continuous query
    template : string
        The name of the template
    session : requests.Session
        The session for the windows

    Parameters
    ----------
    *args : one-or-more arguments, optional
        Positional arguments to MutableMapping
    **kwargs : keyword arguments, optional
        Keyword arguments to MutableMapping

    '''

    def __init__(self, template, *args, **kwargs):
        collections.MutableMapping.__init__(self, *args, **kwargs)
        self._data = dict()
        self.template = template
        self.project = None
        self.project_handle = None
        self.contquery = None
        self.session = None

    @property
    def session(self):
        '''
        The session for the windows

        Returns
        -------
        string

        '''
        return self._session

    @session.setter
    def session(self, value):
        self._session = value
        for item in self._data.values():
            item.session = self._session

    @property
    def project(self):
        '''
        The project that windows are associated with

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        self._project = getattr(value, 'name', value)
        for item in self._data.values():
            item.project = self._project

    @property
    def contquery(self):
        '''
        The continuous query that windows are associated with

        Returns
        -------
        string

        '''
        return self._contquery

    @contquery.setter
    def contquery(self, value):
        self._contquery = getattr(value, 'name', value)
        for item in self._data.values():
            item.contquery = self._contquery
        if hasattr(value, 'project'):
            self.project = value.project

    @property
    def template(self):
        '''
        The tempalte that windows are associated with

        Returns
        -------
        string

        '''
        return self._template

    @template.setter
    def template(self, value):
        self._template = value
        for item in self._data.values():
            item.template = self._template
        if hasattr(value, 'contquery'):
            self.contquery = value.contquery
        if hasattr(value, 'project'):
            self.project = value.project

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        if not isinstance(value, BaseWindow):
            raise TypeError('Only Window objects can be values '
                            'in a template')

        if key in self._data.keys():
            oldkey = key
            suffix = 0
            while key in self._data.keys():
                key = key.strip(str(suffix))
                suffix += 1
                key = key + str(suffix)

            warnings.warn('%s already exists in Template %s, renamed to %s' % (oldkey, self.template.name, key),
                          Warning)

        value._register_to_project(self.template)

        oldname = value.name

        value.base_name = key
        value.project = self.project
        value.contquery = self.contquery
        value.template = self.template
        value.session = self.session
        self._data[key] = value

        # Make sure targets get updated with new name
        if oldname != value.name:
            for window in self._data.values():
                for target in set(window.targets):
                    if target.name == oldname:
                        role, slot = target.role, target.slot
                        window.targets.discard(target)
                        window.add_target(value, role=role, slot=slot)

    def __delitem__(self, key):
        del self._data[key]
        for window in self._data.values():
            window.delete_target(key)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class Template(ESPObject, collections.MutableMapping):
    '''
    ESP Template

    Parameters
    ----------
    name : string
        Name of the continuous query
    trace : string, optional
        One or more space-separated window names or IDs
    index_type : string, optional
        A default index type for all windows in the template that
        do not explicitly specify an index type
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    timing_threshold : int, optional
        When a window in the template takes more than value microseconds to
        compute for a given event or event block, a warning message is logged
    include_singletons : bool, optional
        Specify whether to add unattached source windows
    description : string, optional
        Description of the template

    Attributes
    ----------
    project : string or Project
        Name of the project the template is associated with
    contquery : string or ContinuousQuerry
        The name of the continuous query
    windows : dict
        Collection of windows in the template
    metadata : dict
        Metadata dictionary
    url : string
        URL of the template

    Notes
    -----
    All parameters are also available as instance attributes.

    Returns
    -------
    :class:`Template`

    '''
    trace = attribute('trace', dtype='string')

    index_type = attribute('index', dtype='string',
                           values={'rbtree': 'pi_RBTREE', 'hash': 'pi_HASH',
                                   'ln_hash': 'pi_LN_HASH', 'cl_hash': 'pi_CL_HASH',
                                   'fw_hash': 'pi_FW_HASH', 'empty': 'pi_EMPTY'})
    timing_threshold = attribute('timing-threshold', dtype='int')
    include_singletons = attribute('include-singletons', dtype='bool')

    file_dir = os.path.split(__file__)[0]
    template_list = collections.defaultdict(dict)

    def __init__(self, name, trace=None, index_type=None,
                 timing_threshold=None, include_singletons=None, description=None):
        self.windows = WindowDict(self)
        ESPObject.__init__(self, attrs=locals())
        self.project = None
        self.contquery = None
        self.tag = None
        self.name = name or gen_name(prefix='tp_')
        self.description = description
        self.mas_modules = []
        self.metadata = {}
        self.input_windows = list()
        self.output_windows = list()
        self.required_parameter_map = collections.defaultdict(list)

    @property
    def session(self):
        '''
        The requests.Session object for the continuous query

        Returns
        -------
        string

        '''
        return ESPObject.session.fget(self)

    @session.setter
    def session(self, value):
        ESPObject.session.fset(self, value)
        self.windows.session = value

    @property
    def name(self):
        '''
        The name of the template

        Returns
        -------
        string

        '''
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def contquery(self):
        '''
        The name of the continuous query

        Returns
        -------
        string

        '''
        return self._contquery

    @contquery.setter
    def contquery(self, value):
        self._contquery = getattr(value, 'name', value)
        self.windows.contquery = self._contquery

    @property
    def project(self):
        '''
        The name of the project

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        self._project = getattr(value, 'name', value)
        self.windows.project = self._project

    @property
    def input_windows(self):
        return ','.join(self._input_windows)

    @input_windows.setter
    def input_windows(self, value):
        if not isinstance(value, list):
            raise TypeError('Please use a list to set input_windows')

        self._input_windows = list()
        for window in value:
            self.add_input_windows(window)

    def add_input_windows(self, *windows):
        '''
        Add input_windows

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to add as input_windows for Template

        '''

        for window in windows:
            base_name = getattr(window, 'base_name', window)
            if not base_name or base_name not in self.windows:
                raise KeyError('%s is not a one of Template %s' %
                               (base_name, self.name))
            elif base_name in self._input_windows:
                print('%s is already a one of input_windows' % base_name)
            else:
                self._input_windows.append(base_name)

        return self._input_windows

    def delete_input_windows(self, *windows):
        '''
        Delete input_windows

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to deleted from input_windows

        '''

        for window in windows:
            base_name = getattr(window, 'base_name', window)
            if base_name not in self.windows:
                raise KeyError('%s is not a one of Template %s' %
                               (base_name, self.name))
            elif base_name not in self._input_windows:
                print('%s is not a one of input_windows' % base_name)
            else:
                self._input_windows.remove(base_name)

        return self._input_windows

    @property
    def output_windows(self):
        return ','.join(self._output_windows)

    @output_windows.setter
    def output_windows(self, value):
        if not isinstance(value, list):
            raise TypeError('Please use a list to set output_windows')

        self._output_windows = list()
        for window in value:
            self.add_output_windows(window)

    def add_output_windows(self, *windows):
        '''
        Add output_windows

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to add as output_windows for Template

        '''

        for window in set(windows):
            base_name = getattr(window, 'base_name', window)
            if base_name not in self.windows:
                raise KeyError('%s is not a one of Template %s' %
                               (base_name, self.name))
            elif base_name in self._output_windows:
                print('%s is already a one of output_windows' % base_name)
            else:
                self._output_windows.append(base_name)

        return self._output_windows

    def delete_output_windows(self, *windows):
        '''
        Delete output_windows

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to deleted from output_windows

        '''

        for window in set(windows):
            base_name = getattr(window, 'base_name', window)
            if base_name not in self.windows:
                raise KeyError('%s is not a one of Template %s' %
                               (base_name, self.name))
            elif base_name not in self._output_windows:
                print('%s is not a one of input_windows' % base_name)
            else:
                self._output_windows.remove(base_name)

        return self._output_windows

    def create_mas_module(self, language, module, func_names, mas_store=None,
                          mas_store_version=None, description=None,
                          code_file=None, code=None):
        '''
        Create a MAS module object

        Parameters
        ----------
        language : string
            The name of the programming language
        module : string, optional
            Name of the MAS module
        func_names : string or list-of-strings, optional
            The function names exported by the module

        Returns
        -------
        :class:`MASModule`

        '''
        out = MASModule(language, module, func_names, mas_store=mas_store,
                        mas_store_version=mas_store_version,
                        description=description, code_file=code_file,
                        code=code)
        out.project = self.project
        out.session = self.session
        return out

    def set_parameters(self, window, **parameters):
        '''
        Set parameters

        Parameters
        ----------
        window : Window
            The Window object to set parameters
        **parameters : keyword-arguments, optional
            The parameters to set

        '''
        base_name = getattr(window, 'base_name', window)
        try:
            window = self.windows[base_name]
        except KeyError:
            raise KeyError('%s is not a one of Template %s' %
                                         (base_name, self.name))

        if not isinstance(window, (CalculateWindow, TrainWindow, ModelReaderWindow)):
            raise TypeError('Only CalculationWindow and TrainWindow objects support the method')
        return window.set_parameters(**parameters)

    def set_inputs(self, window=None, **input_map):
        '''
        Set inputs

        Parameters
        ----------
        window : Window, optional
            The Window object to set inputs, default value is None
        **input_map : keyword-arguments, optional
            The parameters to set

        '''

        if window is None:
            try:
                window = self._input_windows[0]
                print("INFO: window is not specified, default for first input window %s" % window)
            except IndexError:
                raise IndexError("Please specify input_windows for Template %s first" % self.name)

        base_name = getattr(window, 'base_name', window)
        try:
            window = self.windows[base_name]
        except KeyError:
            raise ValueError('%s is not a one of Template %s' %
                             (base_name, self.name))

        if not isinstance(window, (TrainWindow, CalculateWindow, ScoreWindow)):
            raise TypeError('Only CalculationWindow, TrainWindow and ScoreWindow objects support the method')
        return window.set_inputs(**input_map)

    def set_outputs(self, window=None, **output_map):
        '''
        Set outputs

        Parameters
        ----------
        window : Window, optional
            The Window object to set outputs, default value is None
        **output_map : keyword-arguments, optional
            The parameters to set

        '''
        if window is None:
            try:
                window = self._output_windows[0]
                print("INFO: window is not specified, default for first output window %s" % window)
            except IndexError:
                raise IndexError("Please specify output_windows for Template %s first" % self.name)

        base_name = getattr(window, 'base_name', window)
        try:
            window = self.windows[base_name]
        except KeyError:
            raise ValueError('%s is not a one of Template %s' %
                             (base_name, self.name))

        if not isinstance(window, (CalculateWindow, ScoreWindow)):
            raise TypeError('Only CalculationWindow and ScoreWindow objects support the method')
        return window.set_outputs(**output_map)

    def add_target(self, obj, **kwargs):
        '''
        Add target for Template

        Parameters
        ----------
        obj : Window or Template
            The Window or Template object to add as targets
        role : string, optional
            The role of the connection
        slot : string, optional
            Indicates the slot number to use from the splitting
            function for the window.

        Returns
        -------
        ``self``

        '''
        try:
            window_name = self._output_windows[0]
        except IndexError:
            raise IndexError("Please specify output_windows for Template %s first" % self.name)

        window = self.windows[window_name]
        if isinstance(obj, BaseWindow):
            window.add_targets(obj, **kwargs)

        elif isinstance(obj, Template):
            try:
                target_window_name = obj._input_windows[0]
            except IndexError:
                raise IndexError("Please specify input_windows for Template %s first" % obj.name)
            window.add_targets(obj.windows[target_window_name], **kwargs)

    def delete_target(self, *objs):
        '''
        Delete targets for Template

        Parameters
        ----------
        obj : Window or Template
            The Window or Template object to deleted from targets

        Returns
        -------
        ``self``

        '''
        try:
            window_name = self._output_windows[0]
        except IndexError:
            raise IndexError("There is no output_windows for Template %s" % self.name)

        window = self.windows[window_name]
        target_set = set(target.name for target in window.targets)
        for obj in objs:
            if isinstance(obj, BaseWindow):
                window.delete_targets(obj)

            elif isinstance(obj, Template):
                for possible_target in obj._input_windows:
                    if obj.windows[possible_target].name in target_set:
                        window.delete_targets(obj.windows[possible_target])

    def add_window(self, window):
        '''
        Add a window to the template

        Parameters
        ----------
        window : Window
            The Window object to add

        Returns
        -------
        :class:`Window`

        '''
        if not window.base_name:
            window.base_name = gen_name(prefix='w_')

        self.windows[window.base_name] = window

        return window

    def add_windows(self, *windows):
        '''
        Add one or more windows to the template

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to add

        Returns
        -------
        tuple of :class:`Window`s

        '''
        for item in windows:
            self.add_window(item)
        return windows

    def import_template(self, template, internal_only=True):
        '''
        import a template object

        Parameters
        ----------
        template : Template
            A Template object to be imported to current template
        internal_only: bool, optional
            Only includes the internal edges or not, default value is True

        Returns
        -------
        :class:`Template`
        '''
        if template is self:
            warnings.warn('You are importing the self template.', Warning)
            return self.import_template(self.copy(None), internal_only=internal_only)

        ref_dict = {}
        for key, window in sorted(six.iteritems(template.windows)):
            copied_win = window.copy(deep=True)
            self.windows[key] = copied_win
            ref_dict[key] = copied_win.base_name

        for old_name, new_name in ref_dict.items():
            copied_win = self.windows[new_name]
            win = template.windows[old_name]
            copied_win.targets = set()
            for target in set(win.targets):
                if target.base_name in ref_dict:
                    copied_win.targets.add(Target(name=ref_dict[target.base_name], template=self,
                                                  role=target.role, slot=target.slot))
                elif not internal_only:
                    copied_win.targets.add(target)

        return

    def copy(self, name, deep=True, internal_only=True):
        '''
        Return a copy of the template

        Parameters
        ----------
        name : string
            Name of the copied template
        deep : bool, optional
            Copy the sub-objects or not, default value is True
        internal_only: bool, optional
            Only includes the internal edges or not, default value is True

        Returns
        -------
        :class:`Template`

        '''
        out = type(self)(name)

        out.session = self.session
        out.contquery = self.contquery
        out.project = self.project

        for key, value in self._get_attributes(use_xml_values=False).items():
            if key != 'name':   # do NOT copy the old name
                setattr(out, key, value)

        if deep:
            for k, win in self.windows.items():
                out.windows[k] = win.copy(deep=deep)
                out.windows[k].targets = set()
                for target in set(win.targets):
                    if target.template.name == self.name:
                        out.windows[k].targets.add(Target(name=target.base_name, template=out,
                                                          role=target.role, slot=target.slot))
                    elif not internal_only:
                        out.windows[k].targets.add(target)
        else:
            out.windows.update(self.windows)

        out.input_windows = self._input_windows
        out.output_windows = self._output_windows
        out.required_parameter_map = self.required_parameter_map

        return out

    def __copy__(self):
        return self.copy(name=None, deep=False)

    def __deepcopy__(self, memo):
        return self.copy(name=None, deep=True)

    @property
    def fullname(self):
        return '%s.%s.%s' % (self.project, self.contquery, self.name)

    @property
    def url(self):
        '''
        URL of the Template

        Returns
        -------
        string

        '''
        self._verify_project()
        return urllib.parse.urljoin(self.base_url, '%s/%s/%s/' %
                                    (self.project, self.contquery, self.name))

    @classmethod
    def from_xml(cls, data, template_name, tag=None, contquery=None, project=None, session=None):
        '''
        Create template from XML definition

        Parameters
        ----------

        data : xml-string or ElementTree.Element
            XML template definition
        template_name: string
            The name for the newly created Template object

        tag: string, optional
            Type of imported template
        contquery : string, optional
            The name of Continuous Query
        project : string, optional
            The name of Project
        session : requests.Session, optionals
            The session object

        Returns
        -------
        :class:`Template`

        '''
        out = cls(template_name)
        out.session = session
        out.project = project
        out.contquery = contquery

        if isinstance(data, six.string_types):
            if re.match(r'^\s*<', data):
                data = data
            elif os.path.isfile(data):
                data = open(data, 'r').read()
            else:
                data = urllib.request.urlopen(data).read().decode('utf-8')
            data = xml.from_xml(data)

        try:
            del data.attrib['name']
        except:
            pass

        try:
            out.tag = data.attrib['tag']
        except:
            out.tag = tag

        out._set_attributes(data.attrib)

        for desc in data.findall('./description'):
            out.description = desc.text

        for item in data.findall('./mas-modules/mas-module'):
            out.mas_modules.append(MASModule.from_xml(item, session=session))

        for item in data.findall('./windows/*'):
            try:
                wcls = get_window_class(item.tag)
            except KeyError:
                raise TypeError('Unknown window type: %s' % item.tag)

            window = wcls.from_xml(item, session=session)
            out.windows[window.base_name] = window

        for item in data.findall('./edges/*'):
            for target in re.split(r'\s+', item.attrib.get('target', '').strip()):
                if not target or target not in out.windows:
                    continue
                for source in re.split(r'\s+', item.attrib.get('source', '').strip()):
                    if not source or source not in out.windows:
                        continue
                    out.windows[source].add_target(out.windows[target], role=item.get('role'),
                                                   slot=item.get('slot'))
        try:
            out.input_windows = re.sub("[^\w]", " ", data.attrib['input-windows']).split()
        except:
            pass

        try:
            out.output_windows = re.sub("[^\w]", " ", data.attrib['output-windows']).split()
        except:
            pass

        for item in data.findall('./metadata/meta'):
            if 'id' in item.attrib.keys():
                out.metadata[item.attrib['id']] = item.text
            elif 'name' in item.attrib.keys():
                out.metadata[item.attrib['name']] = item.text

        for item in data.findall('./required-parameter-map/properties/property'):
            field = item.text.split(',')
            out.required_parameter_map[item.attrib['name']] = field

        return out

    from_element = from_xml

    def to_element(self, template=False):
        '''
        Export template definition to ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''

        extra_attrs = [item for item in ['tag', 'input_windows', 'output_windows'] if getattr(self, item)]
        out = xml.new_elem('template', xml.get_attrs(self, exclude=['name', 'project', 'contquery'], extra=extra_attrs))

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        if self.metadata:
            metadata = xml.add_elem(out, 'metadata')
            for key, value in sorted(six.iteritems(self.metadata)):
                xml.add_elem(metadata, 'meta', attrib=dict(id=key),
                             text_content=value)

        if self.mas_modules:
            mods = xml.add_elem(out, 'mas-modules')
            for item in self.mas_modules:
                xml.add_elem(mods, item.to_element())

        windows = xml.add_elem(out, 'windows')

        sources = {}
        if self.windows:
            edges = []
            for name, window in sorted(six.iteritems(self.windows)):
                win_out = window.to_element()
                if not template:
                    win_out.attrib['name'] = window.base_name
                xml.add_elem(windows, win_out)
                for target in window.targets:
                    sources.setdefault(target.name, []).append(window.name)
                    attrib = dict(source=window.base_name, target=target.base_name)
                    if target.role:
                        attrib['role'] = target.role
                    if target.slot:
                        attrib['slot'] = target.slot
                    edges.append((target._index, attrib))
            if edges:
                elem = xml.add_elem(out, 'edges')
                for i, attrib in sorted(edges):
                    xml.add_elem(elem, 'edge', attrib=attrib)

            if self.required_parameter_map:
                mappings = xml.add_elem(out, 'required-parameter-map')
                xml.add_properties(mappings, self.required_parameter_map, bool_as_int=True)

        # Replace "inherit" data types with the real data type
        n_inherit = -1
        while True:
            inherit = out.findall('./windows/*/schema/fields/field[@type="inherit"]')
            if len(inherit) == n_inherit:
                break
            n_inherit = len(inherit)
            for window in out.findall('./windows/*'):
                for field in window.findall('./schema/fields/field[@type="inherit"]'):
                    for source in sources[window.attrib['name']]:
                        fname = field.attrib['name']
                        if source not in self.windows:
                            raise ValueError("Could not determine data type of "
                                             "field '%s' on window '%s'" % (fname, source))
                        win = self.windows[source]
                        if hasattr(win, 'schema') and fname in win.schema:
                            dtype = win.schema[fname].type
                            field.set('type', dtype)

        # return windows, elem
        return out

    def to_xml(self, pretty=False, template=False):
        '''
        Export template definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output embed whitespaced for readability or not, default value is False
        template : bool, optional
            To include template name or not, default value is False

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(template=template), pretty=pretty)

    def save_xml(self, dest, mode='w', pretty=True, **kwargs):
        '''
        Save the template XML to a file

        Parameters
        ----------
        dest : string or file-like
            The destination of the XML content
        mode : string, optional
            The write mode for the output file (only used if `dest` is a string)
        pretty : boolean, optional
            Should the XML include whitespace for readability or not, default value is True

        '''
        if isinstance(dest, six.string_types):
            with open(dest, mode=mode, **kwargs) as output:
                output.write(self.to_xml(pretty=pretty))
        else:
            dest.write(self.to_xml(pretty=pretty))

    def export_to(self, type='xml', pretty=False):
        if type == 'xml':
            return self.to_xml(pretty=pretty)

        return

    def import_from(self, type='xml'):
        if type == 'xml':
            return self.from_xml()

        return

    def _persist_metadata(self):
        if self.metadata:
            self._set_metadata(self.metadata)

    def _clear_metadata(self):
        self.metadata.clear()

    def _set_metadata(self, data):
        for key, value in six.iteritems(data):
            self._put(urllib.parse.urljoin(self.base_url,
                                           'projectMetadata/%s/%s/%s' %
                                           (self.project, self.name, key)),
                      data='%s' % value)

    def _del_metadata(self, *data):
        for key in data:
            self._delete(urllib.parse.urljoin(self.base_url,
                                              'projectMetadata/%s/%s/%s' %
                                              (self.project, self.name, key)))

    def save_xml(self, dest, mode='w', pretty=True, **kwargs):
        '''
        Save the template XML to a file

        Parameters
        ----------
        dest : string or file-like
            The destination of the XML content
        mode : string, optional
            The write mode for the output file (only used if `dest` is a string)
        pretty : boolean, optional
            Should the XML include whitespace for readability or not, default value is True

        '''
        if isinstance(dest, six.string_types):
            with open(dest, mode=mode, **kwargs) as output:
                output.write(self.to_xml(pretty=pretty))
        else:
            dest.write(self.to_xml(pretty=pretty))

    def to_graph(self, graph=None, schema=False, detail=False):
        '''
        Export template definition to graphviz.Digraph

        Parameters
        ----------
        graph : graphviz.Graph, optional
            The parent graph to add to
        schema : bool, optional
            Include window schemas or not, default value is False
        detail : bool, optional
            Show template detail or not, default value is False

        Returns
        -------
        :class:`graphviz.Digraph`

        '''
        try:
            import graphviz as gv
        except ImportError:
            raise ImportError('The graphviz module is required for exporting to graphs.')

        if graph is None:
            graph = gv.Digraph(format='svg')
            graph.attr('node', shape='rect')
            graph.attr('graph', rankdir='LR', center='false')
            graph.attr('edge', fontname='times-italic')

        label = 'Template-%s: ' % self.tag if self.tag else 'Template: '
        label = label + self.name
        if self.windows and detail:
            tgraph = gv.Digraph(format='svg', name='cluster_%s' % self.fullname.replace('.', '_'))
            tgraph.attr('node', shape='rect')
            tgraph.attr('graph', fontname='helvetica')
            tgraph.attr('edge', fontname='times-italic')
            tgraph.attr(label=label, labeljust='l', style='rounded,bold,dashed', color='blue', fontcolor='black')

            for wkey, window in sorted(self.windows.items()):
                window.to_graph(graph=tgraph,
                                schema=schema or get_option('display.show_schema'))
                for target in window.targets:
                    if target.base_name in self.windows and target.template and target.template.name == self.name:
                        graph.edge(window.fullname, self.windows[target.base_name].fullname, label=target.role or '')

            graph.subgraph(tgraph)

        else:
            graph.node(self.fullname, label=label, labeljust='l',
                       style='bold,filled', color='blue', fillcolor='#f0f0f0',
                       fontcolor='blue', margin='.25,.17', fontname='helvetica')

        return graph

    def _repr_svg_(self):
        try:
            return scale_svg(self.to_graph()._repr_svg_())
        except ImportError:
            raise AttributeError('_repr_svg_')

    def __str__(self):
        return '%s(name=%s, project=%s)' % (type(self).__name__,
                                            repr(self.name),
                                            repr(self.project))

    def __repr__(self):
        return str(self)

    def rename_window(self, window, newname):
        '''
        Rename a window and update targets

        Parameters
        ----------
        window : string or Window
            The window to rename
        newname : string
            The new name of the Window object

        '''

        oldname = getattr(window, 'base_name', window)
        self.windows[newname] = self.windows[oldname]
        del self.windows[oldname]

    def delete_windows(self, *windows):
        '''
        Delete windows and update targets

        Parameters
        ----------
        windows : one-or-more strings or Window objects
            The window to delete

        '''
        for item in windows:
            to_delete = self.windows[getattr(item, 'base_name', item)]
            to_delete.template = None
            del self.windows[getattr(item, 'base_name', item)]

    delete_window = delete_windows

    def subscribe(self, mode='streaming', pagesize=50, filter=None,
                  sort=None, interval=None, limit=None, horizon=None, reset=True):
        '''
          Subscribe to events

          Parameters
          ----------
          mode : string, optional
              The mode of subscriber: 'updating' or 'streaming'
          pagesize : int, optional
              The maximum number of events in a page
          filter : string, optional
              Functional filter to subset events
          sort : string, optional
              Sort order for the events (updating mode only)
          interval : int, optional
              Interval between event sends in milliseconds
          limit : int, optional
              The maximum number of rows of data to keep in the internal
              DataFrame object.
          horizon : int or datetime.datetime or string, optional
              Specifies a condition that stops the subscriber.
              If an int, the subscriber stops after than many events.
              If a datetime.datetime, the subscriber stops after the specified
              date and time.  If a string, the string is an expression
              applied to the event using the :meth:`DataFrame.query`
              method.  If that query returns any number of rows, the
              subscriber is stopped.
          reset : bool, optional
              If True, the internal data is reset on subsequent calls
              to the :meth:`subscribe` method.

          See Also
          --------
          :meth:`unsubscribe`
          :class:`Subscriber`

          '''
        for k, win in self.windows.items():
            win.subscribe(mode, pagesize, filter, sort, interval, limit, horizon, reset)

    def unsubscribe(self):
        '''
        Stop event processing

        See Also
        --------
        :meth:`subscribe`

        '''
        for k, win in self.windows.items():
            win.unsubscribe()

#
# MutableMapping methods
#

    def __getitem__(self, key):
        return self.windows[key]

    def __setitem__(self, key, value):
        self.windows[key] = value

    def __delitem__(self, key):
        del self.windows[key]

    def __iter__(self):
        return iter(self.windows)

    def __len__(self):
        return len(self.windows)

    def __contains__(self, value):
        return value in self.windows





from xml.etree import ElementTree
import pandas as pd
import esppy.espapi.tools as tools
import threading
import logging
import esppy
import time
import six
import re
import os

if os.getenv("ESPPY_LOG") != None:
    logging.basicConfig(filename=os.getenv("ESPPY_LOG"),level=logging.INFO)

class Connection(tools.Options):
    def __init__(self,host,port,secure,**kwargs):
        tools.Options.__init__(self,**kwargs)
        self._host = host
        self._port = port
        self._secure = secure
        self._websocket = None
        self._handshakeComplete = False
        self._headers = None
        self._authorization = None

    def start(self,readyCb = None):
        if (self.isConnected):
            return

        self.clear()

        url = self.getUrl()

        if (url == None):
            raise Exception("invalid url")

        ws = esppy.websocket.WebSocketClient(url,on_message=self.on_message,on_error=self.on_error,on_open=self.on_open,on_close=self.on_close)
        ws.connect()

    def stop(self):
        if (self.isConnected):
            self.clear()
            return(True)
        return(False)

    def restart(self):
        self.clear()
        self.start()

    def send(self,data):
        if self._websocket != None:
            #logging.info("SEND: " + str(data))
            self._websocket.send(str(data))

    def getUrl(self):
        return(None)

    def message(self,message):
        if (self.isHandshakeComplete):
            return

        name = ""
        value = None

        for i in range(0,len(message)):
            c = message[i]
            if (c == '\n'):
                if (len(name) == 0):
                    break
                if (self._headers == None):
                    self._headers = {}
                if (value != None):
                    self._headers[name] = value.strip()
                else:
                    self._headers[name] = ""

                name = ""
                value = None
            elif (value != None):
                value += c
            elif (c == ':'):
                value = ""
            else:
                name += c

        status = self.getHeader("status")

        if (status != None):
            value = int(status)
            if (value == 200):
                self._handshakeComplete = True
                self.handshakeComplete()

                if (tools.supports(self._delegate,"connected")):
                    self._delegate.connected(self)
            elif (value == 401):
                if (self._authorization != None):
                    self._websocket.send(self._authorization)
                elif (tools.supports(self._delegate,"authenticate")):
                    scheme = self.getHeader("www-authenticate")
                    self._delegate.authenticate(self,scheme)

    def close(self):
        pass

    def error(self):
        pass

    def closed(self):
        pass

    def on_open(self,ws):
        self._websocket = ws

    def on_close(self,ws,code,reason):
        self.clear()
        self.closed()

    def on_error(self,ws,error):
        self.clear()
        self.error()

    def on_message(self,ws,message):
        self.message(message)

    def setAuthorization(self,value):
        self._authorization = value

        if self.isConnected and self.isHandshakeComplete == False:
            self._websocket.send(self._authorization)

    def clear(self):
        if (self._websocket != None):
            self._websocket.close()
            self._websocket = None

        self._handshakeComplete = False
        self._headers = None

    def getHeader(self,name):
        value = None
        if (self._headers != None):
            if name in self._headers:
                value = self._headers[name]
        return(value)

    def getHost(self):
        return(self._host)

    def getPort(self):
        return(self._port)

    def getProtocol(self):
        if self._secure:
            return("wss")
        else:
            return("ws")

    def getHttpProtocol(self):
        if self._secure:
            return("https")
        else:
            return("http")

    def isSecure(self):
        return(self._secure)

    def handshakeComplete(self):
        pass

    @property
    def isConnected(self):
        return(self._websocket != None)

    @property
    def isHandshakeComplete(self):
        return(self._handshakeComplete)

class ServerConnection(Connection):

    _windowClasses = {
        "source":"input",
        "filter":"transformation",
        "aggregate":"transformation",
        "compute":"transformation",
        "union":"transformation",
        "join":"transformation",
        "copy":"transformation",
        "functional":"transformation",
        "notification":"utility",
        "pattern":"utility",
        "counter":"utility",
        "geofence":"utility",
        "procedural":"utility",
        "model-supervisor":"analytics",
        "model-reader":"analytics",
        "train":"analytics",
        "calculate":"analytics",
        "score":"analytics",
        "text-context":"textanalytics",
        "text-category":"textanalytics",
        "text-sentiment":"textanalytics",
        "text-topic":"textanalytics"
    }

    def __init__(self,host,port,secure,delegate,**kwargs):
        Connection.__init__(self,host,port,secure,**kwargs)
        self._delegate = delegate
        self._datasources = {}
        self._publishers = {}
        self._stats = Stats(self)
        self._log = Log(self)
        self._modelDelegates = {}
        self._urlPublishers = {}
        self._autoReconnect = True

    def message(self,message):
        if self.isHandshakeComplete == False:
            Connection.message(self,message)
            return

        #logging.info("MSG: " + message)

        xml = ElementTree.fromstring(str(message))

        if xml.tag == "events" or xml.tag == "schema" or xml.tag == "info":

            datasource = None

            if "id" in xml.attrib:
                id = xml.get("id")
                if id in self._datasources:
                    datasource = self._datasources[id]
            if datasource != None:
                if xml.tag == "events":
                    datasource.events(xml)
                elif xml.tag == "schema":
                    datasource.setSchema(xml)
                elif xml.tag == "info":
                    datasource.info(xml)
        elif xml.tag == "stats":
            self._stats.process(xml)
        elif xml.tag == "log":
            self._log.process(xml)
        elif (xml.tag == "model"):
            if "id" in xml.attrib:
                id = xml.get("id")

                if id in self._modelDelegates:
                    delegate = self._modelDelegates[id]
                    delegate.deliver(xml)
                    del self._modelDelegates[id]
        elif (xml.tag == "url-publisher"):
            if "id" in xml.attrib:
                id = xml.get("id")
                if id in self._urlPublishers:
                    publisher = self._urlPublishers[id]
                    if "complete" in xml.attrib:
                        complete = xml.get("complete")
                        publisher["complete"] = (complete == "true")
                        del self._urlPublishers[id]

        else:
            print("THE MSG: " + message)

    def getUrl(self):
        url = ""
        url += self.getProtocol()
        url += "://"
        url += self.getHost()
        url += ":"
        url += self.getPort()
        url += "/eventStreamProcessing/v2/connect"
        return(url)

    def getEventCollection(self,path,**kwargs):
        ec = EventCollection(self,path,**kwargs)
        self._datasources[ec._id] = ec
        if self.isHandshakeComplete:
            ec.open()
        return(ec)

    def getEventStream(self,path,**kwargs):
        es = EventStream(self,path,**kwargs)
        self._datasources[es._id] = es
        if self.isHandshakeComplete:
            es.open()
        return(es)

    def getPublisher(self,path,**kwargs):
        publisher = Publisher(self,path,**kwargs);
        self._publishers[publisher._id] = publisher;
        if self.isHandshakeComplete:
            publisher.open();
        return(publisher);

    def publishUrl(self,path,url,**kwargs):
        opts = tools.Options(**kwargs)
        blocksize = opts.getOpt("blocksize",1)
        wait = opts.getOpt("wait",False)

        id = tools.guid();

        o = {}
        o["request"] = "url-publisher"
        o["id"] = id
        o["window"] = path;
        o["url"] = url;
        o["blocksize"] = blocksize

        publisher = {"complete":False}
        self._urlPublishers[id] = publisher

        self.send(o)

        if wait:
            while publisher["complete"] == False:
                time.sleep(1)

    def publishDataFrame(self,path,df,**kwargs):
        opts = tools.Options(**kwargs)

        size = opts.getOpt("size",100)
        blocksize = opts.getOpt("blocksize",1)

        id = tools.guid()

        request = {}
        request["request"] = "publisher"
        request["id"] = id
        request["action"] = "set"
        request["window"] = path
        self.send(request)

        request["action"] = "publish"

        data = []

        for index, row in df.iterrows():
            o = {}
            for col in df.columns:
                o[col] = row[col]
            data.append(o)
            if len(data) == size:
                request["data"] = data
                self.send(request)
                data = []

        if len(data) > 0:
            request["data"] = data
            self.send(request)

        request["data"] = None
        request["action"] = "delete"
        self.send(request)

    def publishList(self,path,l,**kwargs):
        opts = tools.Options(**kwargs)

        size = opts.getOpt("size",100)
        blocksize = opts.getOpt("blocksize",1)

        id = tools.guid()

        request = {}
        request["request"] = "publisher"
        request["id"] = id
        request["action"] = "set"
        request["window"] = path
        self.send(request)

        request["action"] = "publish"

        data = []

        for i in l:
            o = {}
            for key,value in i.items():
                o[key] = i[key]
            data.append(o)
            if len(data) == size:
                request["data"] = data
                self.send(request)
                data = []

        if len(data) > 0:
            request["data"] = data
            self.send(request)

        request["data"] = None
        request["action"] = "delete"
        self.send(request)

    def getStats(self):
        return(self._stats)

    def getLog(self):
        return(self._log)

    def loadModel(self,delegate):
        if tools.supports(delegate,"modelLoaded") == False:
            raise Exception("The stats delegate must implement the modelLoaded method")

        id = tools.guid()
        self._modelDelegates[id] = ModelDelegate(self,delegate)

        o = {}
        o["request"] = "model"
        o["id"] = id
        o["schema"] = True
        o["index"] = True
        o["xml"] = True

        self.send(o)

    def handshakeComplete(self):

        for c in self._datasources.values():
            c.open()

        for p in self._publishers.values():
            p.open()

        if len(self._log._delegates) > 0:
            self._log.start()

        if len(self._stats._delegates) > 0:
            self._stats.set()

        if tools.supports(self._delegate,"connected"):
            self._delegate.connected(self)

    def closed(self):
        for c in self._datasources.values():
            c.clear()

        if tools.supports(self._delegate,"closed"):
            self._delegate.closed(self)

        if self._autoReconnect:
            thread = threading.Thread(target = self.reconnect)
            thread.daemon = True
            thread.start()

    def reconnect(self):
        logging.info("RECONNECT")
        while self.isConnected == False:
            #time.sleep(5)
            #time.sleep(1)
            time.sleep(300)
            try:
                self.start()
            except:
                pass

class Datasource(tools.Options):
    def __init__(self,connection,**kwargs):
        tools.Options.__init__(self,**kwargs)
        self._connection = connection
        self._id = tools.guid()
        self._fields = None
        self._keyFields = None
        self._schema = Schema()
        self._delegates = []
        self._paused = False
        self._data = None

    def setSchema(self,xml):
        self._schema.fromXml(xml)
        for d in self._delegates:
            if tools.supports(d,"schemaSet"):
                d.schemaSet(self)

    def setFilter(self,value):
        self.setOpt("filter",value)
        self.set()

    def getFilter(self):
        return(self.getOpt("filter",""))

    def togglePlay(self):

        code = False

        if self._paused:
            self.play()
            code = True
        else:
            self.pause()

        return(code)

    def getFields(self):
        fields = None
        if self._schema != None:
            fields = self._schema.getFields()
        return(fields)

    def getKeyFields(self):
        fields = None
        if self._schema != None:
            fields = self._schema.getKeyFields()
        return(fields)

    def getKeyFieldNames(self):
        names = []
        fields = self.getKeyFields()
        if fields != None:
            for f in fields:
                names.append(f["name"])
        return(names)

    def getColumnFields(self):
        fields = None
        if self._schema != None:
            fields = self._schema.getColumnFields()
        return(fields)

    def getField(self,name):
        field = None
        if self._schema != None:
            field = self._schema.getField(name)
        return(field)

    def getKey(self,o):
        key = ""

        for f in self._schema._keyFields:
            name = f["name"]
            if (name in o) == False:
                break
            if len(key) > 0:
                key += "-"
            key += str(o[name])

        return(key)

    def getData(self):
        return(self._data)

    def getList(self):
        if isinstance(self._data,dict):
            l = []
            for k,v in self._data.items():
                l.append(v)
            return(l)
        elif isinstance(self._data,list):
            return(self._data)

    def getValues(self,name):

        f = self._schema.getField(name)

        if f == None:
            return(None)

        values = []

        if isinstance(self._data,dict):
            for key,value in self._data.items():
                if name in value:
                    if f["isNumber"]:
                        values.append(float(value[name]))
                    else:
                        values.append(value[name])
                elif f["isNumber"]:
                    values.append(0.0)
                else:
                    values.append("")
        elif isinstance(self._data,list):
            for value in self._data:
                if name in value:
                    if f["isNumber"]:
                        values.append(float(value[name]))
                    else:
                        values.append(value[name])
                elif f["isNumber"]:
                    values.append(0.0)
                else:
                    values.append("")

        return(values)

    def getValuesBy(self,keys,names,delimiter = "."):
        keyFields = []

        for s in keys:
            f = self._schema.getField(s)
            if f == None:
                raise Exception("field " + s + " not found")
            keyFields.append(f)

        timeKeys = False

        if len(keyFields) == 1:
            if keyFields[0]["isDate"]:
                timeKeys = True
            elif keyFields[0]["isTime"]:
                timeKeys = True

        valueFields = []

        for s in names:
            f = self._schema.getField(s)
            if f == None:
                raise Exception("field " + s + " not found")
            valueFields.append(f)

        items = None

        if isinstance(self._data,dict):
            items = self._data.values()
        elif isinstance(self._data,list):
            items = self._data

        if items == None:
            raise Exception("invalid data")

        data = {}

        for o in items:
            key = ""
            for f in keyFields:
                name = f["name"]
                if name in o:
                    if len(key) > 0:
                        key += delimiter
                    key += str(o[name])

            if key in data:
                entry = data[key]
            else:
                entry = {}
                for f in valueFields:
                    name = f["name"]
                    entry[name] = 0.0
                data[key] = entry

            for f in valueFields:
                if f["isNumber"]:
                    name = f["name"]
                    entry[name] += float(o[name])

        keyValues = []
        values = {}

        for f in valueFields:
            name = f["name"]
            values[name] = []

        for k,v in data.items():
            if timeKeys:
                dt = pd.to_datetime(k,unit="us")
                keyValues.append(dt)
            else:
                keyValues.append(k)
            for f in valueFields:
                name = f["name"]
                values[name].append(v[name])

        v = {"keys":keyValues,"values":values}

        return(v)

    def getDataFrame(self,values = None):
        if self._data == None:
            return(None)

        data = {}

        fields = []

        if values != None:
            for v in values:
                f = self._schema.getField(v)
                if f != None:
                    fields.append(f)
        else:
            fields = self._schema._fields

        if isinstance(self._data,dict):
            #data["_key_"] = []

            for f in fields:
                data[f["name"]] = []

            for key,o in self._data.items():
                #data["_key_"].append(key)
                for f in fields:
                    name = f["name"]
                    if name in o:
                        if f["isNumber"]:
                            data[name].append(float(o[name]))
                        else:
                            data[name].append(o[name])
                    elif f["isNumber"]:
                        data[name].append(0.0)
                    else:
                        data[name].append("")
        elif isinstance(self._data,list):

            for f in fields:
                data[f["name"]] = []

            for o in self._data:
                for f in fields:
                    name = f["name"]
                    if name in o:
                        if f["isNumber"]:
                            data[name].append(float(o[name]))
                        else:
                            data[name].append(o[name])
                    elif f["isNumber"]:
                        data[name].append(0.0)
                    else:
                        data[name].append("")

        df = pd.DataFrame(data)

        return(df)

    def getInfo(self):
        return({})

    def addDelegate(self,delegate):
        if tools.supports(delegate,"dataChanged") == False:
            raise Exception("the delegate must implement the dataChanged method")

        tools.addTo(self._delegates,delegate)

    def removeDelegate(self,delegate):
        tools.removeFrom(self._delegates,delegate)

    def clear(self):
        pass

    def deliverDataChange(self,data,clear):
        for d in self._delegates:
            d.dataChanged(self,data,clear)

    def deliverInfoChange(self):
        for d in self._delegates:
            if tools.supports(d,"infoChanged"):
                d.infoChanged(self)

    def isList(self):
        return(self._data != None and isinstance(self._data,list))

    def isDict(self):
        return(self._data != None and isinstance(self._data,dict))

    def handleMessage(self,msg):
        pass

    def events(self,xml):
        pass

    def info(self,xml):
        pass

    @property
    def schema(self):
        return(self._schema)

class EventCollection(Datasource):
    def __init__(self,conn,path,**kwargs):
        Datasource.__init__(self,conn,**kwargs)
        self._path = path
        self._page = 0
        self._pages = 0
        self._data = {}

    def open(self):
        o = {}
        o["request"]= "event-collection"
        o["id"]= self._id
        o["action"]= "set"
        o["window"]= self._path
        o["schema"]= True
        o["info"]= 5
        o["format"]= "xml"

        interval = None

        if self.hasOpt("interval"):
            interval = self.getOpt("interval")
        elif self._connection.hasOpt("interval"):
            interval = self._connection.getOpt("interval")

        if interval == None:
            interval = 1000

        self.setOpt("interval",interval)

        for key,value in self.options.items():
            o[key] = value

        if self.hasOpt("filter") == False:
            o["filter"]= ""

        self._connection.send(o)

    def set(self):
        if self._connection.isHandshakeComplete == False:
            return

        o = {}
        o["request"]= "event-collection"
        o["id"]= self._id
        o["action"]= "set"

        for key,value in self.options.items():
            o[key] = value

        if self.hasOpt("filter") == False:
            o["filter"] = ""

        self._connection.send(o)

    def close(self):
        o = {}
        o["request"] = "event-collection"
        o["id"] = self._id
        o["action"] = "close"
        self._connection.send(o)

    def play(self):
        if self._paused:
            self._paused = False
            o["request"]= "event-collection"
            o["id"]= self._id
            o["action"]= "play"
            self.deliverInfoChange()

    def pause(self):
        if self._paused == False:
            self._paused = True
            o["request"]= "event-collection"
            o["id"]= self._id
            o["action"]= "pause"
            self.deliverInfoChange()

    def handleMessage(self,msg):
        self.loadPage(msg["type"])

    def load(self):
        self.loadPage(None)

    def first(self):
        self.loadPage("first")

    def last(self):
        self.loadPage("last")

    def prev(self):
        self.loadPage("prev")

    def next(self):
        self.loadPage("next")

    def loadPage(self,page):
        o = {}
        o["request"] = "event-collection"
        o["id"] = self._id
        o["action"] = "load"
        if page != None:
            o["page"] = page
        self._connection.send(o)

    def events(self,xml):
        data = []

        nodes = xml.findall("entries/event")

        ub = False

        for n in nodes:
            opcode = n.get("opcode")
            if opcode == None:
                opcode = "insert"

            if opcode == "updateblock":
                ub = True
            elif opcode == "delete":
                if ub:
                    ub = False
                    continue

            o = {}
            o["_opcode"] = opcode

            s = n.get("timestamp")

            if s != None:
                o["_timestamp"] = s

            values = n.findall("./*")

            for v in values:
                datatype = v.get("type")
                content = v.text

                if datatype != None:
                    o[v.tag] = "_data://" + datatype + ":" + content
                else:
                    o[v.tag] = content

            o["_key_"] = self.getKey(o)
            data.append(o)

        clear = "page" in xml.attrib

        self.process(data,clear)

        if clear:
            self.info(xml)

    def info(self,xml):
        if "page" in xml.attrib:
            self._page = int(xml.get("page"))
            self._pages = int(xml.get("pages"))
            self.deliverInfoChange()

    def process(self,events,clear):
        if clear:
            self._data = {}

        for e in events:
            key = e["_key_"]
            if key != None:
                opcode = e["_opcode"]
                if opcode == "delete":
                    if key in self._data:
                        del self._data[key]
                else:
                    o = {}

                    o["_key_"] = key

                    for column in self._schema._columns:
                        if column in e:
                            o[column] = e[column]
                    self._data[key] = o

        self.deliverDataChange(events,clear)

    def getInfo(self):
        info = {}
        info["page"] = self._page
        info["pages"] = self._pages
        return(info)
 
    def getKeyValues(self):
        values = []

        for key,value in self._data.items():
            values.append(key)
        return(values)

    def getTableData(self,values):

        rows = []
        columns = []
        cells = []

        a = []

        if values != None and len(values) > 0:
            for v in values:
                f = self._schema.getField(v)
                if f != None:
                    a.append(f)
                    columns.append(f["name"])
        else:
            for f in self._schema.fields:
                if f["isKey"] == False:
                    a.append(f)
                    columns.append(f["name"])

        for key,value in self._data.items():
            rows.append(key)
            cell = []
            for f in a:
                cell.append(value[f["name"]])
            cells.append(cell)

        return({"rows":rows,"columns":columns,"cells":cells})

    def clear(self):
        self._data = {}
        self.deliverDataChange(None,True)

class EventStream(Datasource):
    def __init__(self,conn,path,**kwargs):
        Datasource.__init__(self,conn,**kwargs)
        self._path = path
        self._data = []
        self._counter = 1

    def open(self):
        o = {}
        o["request"] = "event-stream"
        o["id"] = self._id
        o["action"] = "set"
        o["window"] = self._path
        o["schema"] = True
        o["format"] = "xml"

        interval = None

        if self.hasOpt("interval"):
            interval = self.getOpt("interval")
        elif self._connection.hasOpt("interval"):
            interval = self._connection.getOpt("interval")

        if interval == None:
            interval = 1000

        self.setOpt("interval",interval)

        for n,v in self.options.items():
            o[n] = v

        if self.hasOpt("filter") == False:
            o["filter"]= ""

        self._connection.send(o)

    def set(self):
        o = {}
        o["request"] = "event-stream"
        o["id"] = self._id
        o["action"] = "set"
        o["format"] = "xml"

        for n,v in self.options.items():
            o[n] = v

        if self.hasOpt("filter") == False:
            o["filter"]= ""

        self._connection.send(o)

    def close(self):
        o = {}
        o["request"] = "event-stream"
        o["id"] = self._id
        o["action"] = "close"
        self._connection.send(o)

    def play(self):
        if self._paused:
            self._paused = False
            o["request"]= "event-stream"
            o["id"]= self._id
            o["action"]= "play"
            self.deliverInfoChange()

    def pause(self):
        if self._paused == False:
            self._paused = True
            o["request"]= "event-stream"
            o["id"]= self._id
            o["action"]= "pause"
            self.deliverInfoChange()

    def setSchema(self,xml):
        Datasource.setSchema(self,xml)
        for f in self._schema.fields:
            f["isKey"] = False

        self._keyFields = []

        f = {"name":"_opcode","espType":"utf8str","type":"string","isKey":False,"isNumber":False,"isDate":False,"isTime":False}
        self._schema._fields.insert(0,f)
        self._schema._fieldMap["_opcode"] = f
        self._schema._columns.insert(0,f["name"])

        f = {"name":"_timestamp","espType":"timestamp","type":"date","isKey":False,"isNumber":True,"isDate":False,"isTime":True}
        self._schema._fields.insert(0,f)
        self._schema._fieldMap["_timestamp"] = f
        self._schema._columns.insert(0,f["name"])

        f = {"name":"_counter","espType":"int32","type":"int","isKey":True,"isNumber":True,"isDate":False,"isTime":False}
        self._schema._fields.insert(0,f)
        self._schema._fieldMap["_counter"] = f
        self._schema._columns.insert(0,f["name"])

        self._schema._keyFields = [f]

        self._counter = 1

    def events(self,xml):
        data = []

        nodes = xml.findall("entries/event")

        ignoreDeletes = self.getOpt("ignore_deletes",False)

        for n in nodes:
            opcode = n.get("opcode")
            if opcode == None:
                opcode = "insert"

            if opcode == "delete" and ignoreDeletes:
                continue

            o = {}
            o["_opcode"] = opcode

            s = n.get("timestamp")

            if s != None:
                o["_timestamp"] = s

            values = n.findall("./*")

            for v in values:
                datatype = v.get("type")
                content = v.text

                if datatype != None:
                    o[v.tag] = "_data://" + datatype + ":" + content
                else:
                    o[v.tag] = content

            data.append(o)

        self.process(data)

    def process(self,events):
        for e in events:
            o = {}

            for column in self._schema._columns:
                if column in e:
                    o[column] = e[column]

            o["_counter"] = self._counter
            o["_key_"] = self.getKey(o)
            e["_counter"] = o["_counter"]
            e["_key_"] = o["_key_"]

            self._counter += 1

            self._data.append(o)

        maxEvents = self.getOpt("maxevents",50)

        diff = len(self._data) - maxEvents

        if diff > 0:
            for i in range(0,diff):
                del self._data[0]
 
        self.deliverDataChange(events,False)

    def getData(self):
        return(self._data)

    def getKeyValues(self):
        values = []

        for value in self._data:
            values.append(value["_counter"])
        return(values)

    def getTableData(self,values):

        rows = []
        columns = []
        cells = []

        a = []

        if values != None and len(values) > 0:
            for v in values:
                f = self._schema.getField(v)
                if f != None:
                    a.append(f)
                    columns.append(f["name"])
        else:
            for f in self._schema.fields:
                if f["isKey"] == False:
                    a.append(f)
                    columns.append(f["name"])

        for value in self._data:
            rows.append(value["_counter"])
            cell = []
            for f in a:
                cell.append(value[f["name"]])
            cells.append(cell)

        return({"rows":rows,"columns":columns,"cells":cells})

    def clear(self):
        self._data = []
        self.deliverDataChange(None,True)

class Publisher(tools.Options):
    def __init__(self,connection,path,**kwargs):
        tools.Options.__init__(self,**kwargs)
        self._connection = connection
        self._path = path
        self._id = tools.guid()
        self._data = []

    def open(self):
        o = {}
        o["request"] = "publisher"
        o["id"] = self._id
        o["action"] = "set"
        o["window"] = self._path
        o["schema"] = True
        self._connection.send(o)

    def close(self):
        o = {}
        o["request"] = "publisher"
        o["id"] = self._id
        o["action"] = "delete"
        self._connection.send(o)

    def begin(self):
        self._o = {}

    def set(self,name,value):
        self._o[name] = value

    def end(self):
        if self._o != None:
            self._data.append(self._o)
            self._o = {}

    def add(self,o):
        self._data.append(o)

    def publish(self):
        if len(self._data) > 0:
            o = {}
            o["request"] = "publisher"
            o["id"] = self._id
            o["action"] = "publish"
            o["data"] = self._data
            self._connection.send(o)
            self._data = []

    def publishUrl(self,url,**kwargs):
        self._connection.publishUrl(self._path,url,**kwargs)

class Stats(Datasource):
    def __init__(self,connection,**kwargs):
        Datasource.__init__(self,connection,**kwargs)
        self._delegates = []
        self._data = []
        self._memory = {}

        self._schema.addField({"name":"_key_","espType":"utf8str","type":"string","isKey":True,"isNumber":False,"isDate":False,"isTime":False})
        self._schema.addField({"name":"project","espType":"utf8str","type":"string","isKey":False,"isNumber":False,"isDate":False,"isTime":False})
        self._schema.addField({"name":"contquery","espType":"utf8str","type":"string","isKey":False,"isNumber":False,"isDate":False,"isTime":False})
        self._schema.addField({"name":"window","espType":"utf8str","type":"string","isKey":False,"isNumber":False,"isDate":False,"isTime":False})

        #self._schema.addField({"name":"project","espType":"utf8str","type":"string","isKey":True,"isNumber":False,"isDate":False,"isTime":False})
        #self._schema.addField({"name":"contquery","espType":"utf8str","type":"string","isKey":True,"isNumber":False,"isDate":False,"isTime":False})
        #self._schema.addField({"name":"window","espType":"utf8str","type":"string","isKey":True,"isNumber":False,"isDate":False,"isTime":False})

        self._schema.addField({"name":"cpu","espType":"double","type":"double","isKey":False,"isNumber":True,"isDate":False,"isTime":False})
        self._schema.addField({"name":"interval","espType":"int64","type":"int","isKey":False,"isNumber":True,"isDate":False,"isTime":False})
        self._schema.addField({"name":"count","espType":"int64","type":"int","isKey":False,"isNumber":True,"isDate":False,"isTime":False})

    def sortValue(self,o):
        return(o["cpu"])

    def process(self,xml):
        projects = xml.findall(".//project")

        stats = []

        for p in projects:
            contqueries = p.findall(".//contquery")

            for cq in contqueries:
                windows = cq.findall(".//window")

                for w in windows:
                    o = {}
                    o["project"] = p.get("name")
                    o["contquery"] = cq.get("name")
                    o["window"] = w.get("name")
                    o["cpu"] = float(w.get("cpu"))
                    o["interval"] = float(w.get("interval"))
                    o["count"] = w.get("count") != None and float(w.get("count")) or 0
                    o["_key_"] = o["project"] + "." + o["contquery"] + "." + o["window"]
                    stats.append(o)

        stats.sort(key = self.sortValue,reverse = True)

        nodes = xml.findall(".//server-memory")

        self._memory = {}

        if len(nodes) == 1:
            node = nodes[0].find("system")
            if node != None:
                self._memory["system"] = int(node.text)
            node = nodes[0].find("virtual")
            if node != None:
                self._memory["virtual"] = int(node.text)
            node = nodes[0].find("resident")
            if node != None:
                self._memory["resident"] = int(node.text)

        self._data = stats

        for d in self._delegates:
            d.handleStats(self)

    def setOpts(self,**kwargs):
        tools.Options.setOpts(self,**kwargs)
        if len(self._delegates) > 0:
            self.set()

    def setOpt(self,name,value):
        tools.Options.setOpt(self,name,value)
        if len(self._delegates) > 0:
            self.set()

    def set(self):
        o = {}
        o["request"] = "stats"
        o["action"] = "set"
        o["interval"] = self.getOpt("interval",1)

        o["minCpu"] = self.getOpt("mincpu",5)
        o["counts"] = self.getOpt("counts",False)
        o["config"] = self.getOpt("config",False)
        o["memory"] = self.getOpt("memory",True)
        self._connection.send(o)

    def stop(self):
        o = {}
        o["request"] = "stats"
        o["action"] = "stop"
        self._connection.send(o)

    def addDelegate(self,delegate):
        if tools.supports(delegate,"handleStats") == False:
            raise Exception("the delegate must implement the handleStats method")

        if tools.addTo(self._delegates,delegate):
            if len(self._delegates) == 1:
                self.set()

    def removeDelegate(self,delegate):
        if tools.removeFrom(self._delegates,delegate):
            if len(self._delegates) == 0:
                self.stop()

    def getData(self):
        return(self._data)

    def getMemoryData(self):
        return(self._memory)
 
    def getKeyValues(self):
        values = []
        for o in self._data:
            values.append(o["_key_"])
        return(values)

class Log(object):
    def __init__(self,connection):
        self._connection = connection
        self._delegates = []

    def process(self,xml):
        nodes = xml.findall(".")
        if len(nodes) > 0:
            message = nodes[0].text
        for d in self._delegates:
            d.handleLog(self,message)

    def start(self):
        o = {}
        o["request"] = "logs"
        o["capture"] = True
        self._connection.send(o)

    def stop(self):
        o = {}
        o["request"] = "logs"
        o["capture"] = False
        self._connection.send(o)

    def addDelegate(self,delegate):
        if tools.supports(delegate,"handleLog") == False:
            raise Exception("the delegate must implement the handleLog method")

        if tools.addTo(self._delegates,delegate):
            if len(self._delegates) == 1:
                self.start()

    def removeDelegate(self,delegate):
        if tools.removeFrom(self._delegates,delegate) == True:
            if len(self._delegates.length) == 0:
                self.stop()

class Model(object):

    def __init__(self,xml):
        self._projects = []
        self._contqueries = []
        self._windows = []
        self._sourceWindows = []

        projects = xml.findall(".//project")

        for xml in projects:
            p = xml.get("name")

            project = {}
            project["xml"] = xml
            project["key"] = p
            project["name"] = p
            project["index"] = xml.get("index")

            self._projects.append(project)

            project["_contqueries"] = []

            contqueries = xml.findall(".//contquery")

            for cqXml in contqueries:

                cq = cqXml.get("name")
                contquery = {}
                contquery["name"] = cq
                contquery["key"] = p + "/" + cq
                contquery["index"] = cqXml.get("index")
                project["_contqueries"].append(contquery)
                self._contqueries.append(contquery)

                contquery["windows"] = []
                contquery["edges"] = []

                windows = cqXml.findall(".//windows/*")

                if len(windows) == 0:
                    continue

                for windowXml in windows:
                    win = self.addWindow(project,contquery,windowXml)
                    contquery["windows"].append(win)

                edges = cqXml.findall("./edges/edge")

                for edge in edges:
                    sources = edge.get("source").split(" ")
                    targets = edge.get("target").split(" ")

                    for source in sources:
                        source = source.strip()
                        a = p + "/" + cq + "/" + source
                        aw = self.getWindow(a)
                        if aw == None:
                            continue
                        for target in targets:
                            target = target.strip()

                            if len(target) > 0:
                                z = p + "/" + cq + "/" + target
                                zw = self.getWindow(z)

                                if zw != None:
                                    aw["outgoing"].append(zw)
                                    zw["incoming"].append(aw)
                                    contquery["edges"].append({"a":source,"z":target})

    #print(self._projects)

    def getWindow(self,key):
        return(self.get(key,self._windows))

    def get(self,key,a):
        for i in a:
            if i["key"] == key:
                return(i)
        return(None)

    def addWindow(self,project,contquery,xml):
        name = xml.get("name")
        type = xml.get("type")

        if type == None or len(type) == 0:
            type = xml.tag

        a = type.split("-")

        if len(a) > 1:
            type = ""

            for i in range(1,len(a)):
                if len(type) > 0:
                    type += "-"
                type += a[i]

        win = {}
        win["p"] = project["name"]
        win["cq"] = contquery["name"]
        win["name"] = name
        win["type"] = type
        win["index"] = xml.get("index")
        win["xml"] = xml

        if win["index"] == None:
            win["index"] = contquery["index"]

            if win["index"] == None:
                win["index"] = project["index"]

                if win["index"] == None:
                    win["index"] = "pi_HASH"

        win["key"] = project["name"] + "/" + contquery["name"] + "/" + name

        schema = Schema()
        schema.fromXml(xml)

        win["schema"] = schema

        win["incoming"] = []
        win["outgoing"] = []
        win["cpu"] = 0.0

        win["class"] = ServerConnection._windowClasses.get(win["type"],"unknown")

        if win["type"] == "window-source":
            self._sourceWindows[win["key"]] = True

        self._windows.append(win)

        return(win)

    @property
    def windows(self):
        return(self._windows)

class Schema(object):
    def __init__(self):
        self._fields = []
        self._fieldMap = {}
        self._keyFields = []
        self._columns = []

    def fromWindow(self,window):
        self._fields = []
        self._fieldMap = {}
        self._keyFields = []
        self._columns = []

        for name,value in window.schema.items():
            o = {}
            o["name"] = name
            o["espType"] = value.type
            o["isNumber"] = False
            o["isTime"] = False
            o["isDate"] = False

            if o["espType"] == "utf8str":
                o["type"] = "string"
            elif o["espType"] == "int32" or o["espType"] == "int64":
                o["type"] = "int"
                o["isNumber"] = True
            elif o["espType"] == "double" or o["espType"] == "money":
                o["type"] = "float"
                o["isNumber"] = True
            elif o["espType"] == "date":
                o["type"] = "date"
                o["isDate"] = True
            elif o["espType"] == "timestamp":
                o["type"] = "datetime"
                o["isTime"] = True
            else:
                o["type"] = o["espType"]

            o["isKey"] = value.key

            self._fields.append(o)
            self._columns.append(name)

            self._fieldMap[name] = o

            if o["isKey"]:
                self._keyFields.append(o)

    def fromXml(self,xml):
        self._fields = []
        self._fieldMap = {}
        self._keyFields = []
        self._columns = []

        if xml == None:
            raise Exception("no schema specified")

        fields = xml.findall(".//fields/field")

        for f in fields:
            o = {}

            name = f.get("name")
            o["name"] = name
            o["espType"] = f.get("type")
            o["isNumber"] = False
            o["isTime"] = False
            o["isDate"] = False

            if o["espType"] == "utf8str":
                o["type"] = "string"
            elif o["espType"] == "int32" or o["espType"] == "int64":
                o["type"] = "int"
                o["isNumber"] = True
            elif o["espType"] == "double" or o["espType"] == "money":
                o["type"] = "float"
                o["isNumber"] = True
            elif o["espType"] == "date":
                o["type"] = "date"
                o["isDate"] = True
            elif o["espType"] == "timestamp":
                o["type"] = "datetime"
                o["isTime"] = True
            else:
                o["type"] = o["espType"]

            o["isKey"] = (f.get("key") == "true")

            self._fields.append(o)
            self._columns.append(name)

            self._fieldMap[name] = o

            if o["isKey"]:
                self._keyFields.append(o)

    def addField(self,field):
        name = field["name"]
        if (name in self._fieldMap) == False:
            self._fields.append(field)
            self._columns.append(name)

            self._fieldMap[name] = field

            if field["isKey"]:
                self._keyFields.append(field)

    def getField(self,name):
        if name in self._fieldMap:
            return(self._fieldMap[name])
        return(None)

    def getFields(self):
        return(self._fields)

    def getKeyFields(self):
        return(self._keyFields)

    def getColumnFields(self):
        fields = []
        for f in self._fields:
            if field["isKey"] == False:
                keys.append(f)

        return(keys)

    def getFieldType(self,name):
        type = "string"
        if name in self._fieldMap:
            type = self._fieldMap[name]["type"]
        return(type)

    def isNumericField(self,field):
        code = False
        f = None
        if type(field) is str:
            if field in self._fieldMap:
                f = self._fieldMap[field]
        else:
            f = field

        if f != None:
            code = f["isNumber"]

        return(code)

    def isDateField(self,field):
        code = False
        f = None
        if type(field) is str:
            if field in self._fieldMap:
                f = self._fieldMap[field]
        else:
            f = field

        if f != None:
            code = f["isDate"]

        return(code)

    def isTimeField(self,field):
        code = False
        f = None
        if type(field) is str:
            if field in self._fieldMap:
                f = self._fieldMap[field]
        else:
            f = field

        if f != None:
            code = f["isTime"]

        return(code)

    def toXml(self):
        e = ElementTree.Element("schema")
        for field in self._fields:
            f = ElementTree.SubElement(e,"field")
            f.attrib["name"] = field["name"]
            f.attrib["espType"] = field["espType"]
            f.attrib["type"] = field["type"]
            if field["isKey"]:
                f.attrib["isKey"] = "true"
        return(e)

    def toJson(self):
        e = ElementTree.Element("schema")
        fields = []
        for field in self._fields:
            f = {}
            f["name"] = field["name"]
            f["espType"] = field["espType"]
            f["type"] = field["type"]
            f["isNumber"] = field["isNumber"]
            f["isDate"] = field["isDate"]
            f["isTime"] = field["isTime"]
            if field["isKey"]:
                f["isKey"] = "true"
            else:
                f["isKey"] = "false"
            fields.append(f)
        return(fields)

    def __str__(self):
        o = self.toJson()
        return(str(o))

    def toString(self):
        s = ""
        i = 0
        for field in self._fields:
            if field["isKey"] == False:
                continue
            if i > 0:
                s += ","
            s += field["name"]
            s += ":"
            s += field["espType"]
            s += "*"
            i += 1

        for field in self._fields:
            if field["isKey"] == True:
                continue
            if i > 0:
                s += ","
            s += field["name"]
            s += ":"
            s += field["espType"]
            i += 1

        return(s)

    def hasFields(self):
        return(len(self._fields) > 0)

    @property
    def fields(self):
        return(self._fields)

    @property
    def columns(self):
        return(self._columns)

class ModelDelegate(object):

    def __init__(self,connection,delegate):
        self._connection = connection
        self._delegate = delegate

    def deliver(self,xml):
        model = Model(xml)

        if tools.supports(self._delegate,"modelLoaded"):
            self._delegate.modelLoaded(model,self._connection)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP MQTT Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class MQTTSubscriber(Connector):
    '''
    Subscribe to MQ Telemetry Transport (MQTT)

    Parameters
    ----------
    mqtthost : string
        Specifies the MQTT server host name
    mqttclientid : string
        Specifies the string to use as the MQTT Client ID. If empty, a random
        client ID is generated. If NULL, mqttdonotcleansession must be false.
        Must be unique among all clients connected to the MQTT server.
    mqtttopic : string
        Specifies the string to use as an MQTT topic to publish events to.
    mqttqos : string
        Specifies the requested Quality of Service. Values can be 0, 1 or 2.
    mqttmsgtype : string
        Specifies binary, CSV, JSON, or the name of a string field in the
        subscribed window schema.
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    mqttuserid : string, optional
        Specifies the user name required to authenticate the connector's
        session with the MQTT server. 
    mqttpassword : string, optional
        Specifies the password associated with mqttuserid
    mqttport : int, optional
        Specifies the MQTT server port. Default is 1883.
    mqttretainmsg : boolean, optional
        Sets to true to make the published message retained in the MQTT
        Server. Default is false.
    mqttdonotcleansession : boolean, optional
        Instructs the MQTT Server to keep all messages and subscriptions on
        disconnect, instead of keeping them. Default is false.
    mqttkeepaliveinterval : int, optional
        Specifies the number of seconds after which the broker should send a
        PING message to the client if no other messages have been exchanged
        in that time. Default is 10.
    mqttmsgmaxdeliveryattempts : int, optional
        Specifies the number of times the connector tries to resend the
        message in case of failure. Default is 20.
    mqttmsgdelaydeliveryattempts : int, optional
        Specifies the delay in milliseconds between delivery attempts
        specified with mqttmsgmaxdeliveryattempts. Default is 500.
    mqttmsgwaitbeforeretry : int, optional
        Specifies the number of seconds to wait before retrying to send
        messages to the MQTT broker. This applies to publish messages
        with QoS > 0. Default is 20.
    mqttmaxinflightmsg : int, optional
        Specifies the number of QoS 1 and 2 messages that can be simultaneously
        in flight. Default is 20.
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber output
        publishable. The default value is disabled.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received by
        a subscriber that were introduced by a window retention policy.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted as
        an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition used to convert event blocks to protobuf messages.
        When you specify this parameter, you must also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the .proto
        file that you specified with the protofile parameter. Event blocks
        are converted into this message.
    mqttssl : boolean, optional
        Specifies to use SSL/TLS to connect to the MQTT broker. Default is
        false. In order to use SSL/TLS, the ESP encryption overlay
        must be installed.
    mqttsslcafile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM encoded
        trusted CA certificate files. Either mqttsslcafile or mqttsslcapath
        must be specified.
    mqttsslcapath : string, optional
        If mqttssl=true, specifies the path to a directory containing the
        PEM encoded trusted CA certificate files. See mosquitto.conf for
        more details about configuring this directory. Either mqttsslcafile
        or mqttsslcapath must be specified.
    mqttsslcertfile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM
        encoded certificate file for this client. Both mqttsslcertfile and
        mqttsslkeyfile must be provided if one of them is.
    mqttsslkeyfile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM
        encoded private key for this client. Both mqttsslcertfile and
        mqttsslkeyfile must be provided if one of them is.
    mqttsslpassword : boolean, optional
        If mqttssl=true, and if key file is encrypted, specifies the
        password for decryption.
    csvincludeschema : string, optional
        Specifies "never", "once", or "pereventblock". The default value is
        "never". When mqttmsgtype = CSV, prepend output CSV with the window's
        serialized schema.
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].
    mqttpasswordencrypted : boolean, optional
        Specifies that mqttpassword is encrypted
    addcsvopcode : boolean, optional
        Prepends an opcode and comma to input CSV events. The opcode is Insert
        unless publishwithupsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events (with a comma).
        Valid values are "normal" and "partialupdate".
    csvmsgperevent : int, optional
        For CSV, specifies to send one message per event. The default is one
        message per transactional event block or else one message per event.
    csvmsgpereventblock : int, optional
        For CSV, specifies to send one message per event block. The default
        is one message per transactional event block or else one message per event.

    Returns
    -------
    :class:`MQTTSubscriber`

    '''
    connector_key = dict(cls='mqtt', type='subscribe')

    property_defs = dict(
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        mqtthost=prop('mqtthost', dtype='string', required=True),
        mqttclientid=prop('mqttclientid', dtype='string', required=True),
        mqtttopic=prop('mqtttopic', dtype='string', required=True),
        mqttqos=prop('mqttqos', dtype='string', required=True),
        mqttmsgtype=prop('mqttmsgtype', dtype='string', required=True),
        mqttuserid=prop('mqttuserid', dtype='string'),
        mqttpassword=prop('mqttpassword', dtype='string'),
        mqttport=prop('mqttport', dtype='int'),
        mqttretainmsg=prop('mqttretainmsg', dtype='boolean'),
        mqttdonotcleansession=prop('mqttdonotcleansession', dtype='boolean'),
        mqttkeepaliveinterval=prop('mqttkeepaliveinterval', dtype='int'),
        mqttmsgmaxdeliveryattempts=prop('mqttmsgmaxdeliveryattempts', dtype='int'),
        mqttmsgdelaydeliveryattempts=prop('mqttmsgdelaydeliveryattempts', dtype='int'),
        mqttmsgwaitbeforeretry=prop('mqttmsgwaitbeforeretry', dtype='int'),
        mqttmaxinflightmsg=prop('mqttmaxinflightmsg', dtype='int'),
        collapse=prop('collapse', dtype='string'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        mqttssl=prop('mqttssl', dtype='boolean'),
        mqttsslcafile=prop('mqttsslcafile', dtype='string'),
        mqttsslcapath=prop('mqttsslcapath', dtype='string'),
        mqttsslcertfile=prop('mqttsslcertfile', dtype='string'),
        mqttsslkeyfile=prop('mqttsslkeyfile', dtype='string'),
        mqttsslpassword=prop('mqttsslpassword', dtype='string'),
        csvincludeschema=prop('csvincludeschema', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        mqttpasswordencrypted=prop('mqttpasswordencrypted', dtype='boolean'),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='string'),
        csvmsgperevent=prop('csvmsgperevent', dtype='int'),
        csvmsgpereventblock=prop('csvmsgpereventblock', dtype='int') 
    )

    def __init__(self, mqtthost=None, mqttclientid=None, mqtttopic=None,
                 mqttqos=None, mqttmsgtype=None,
                 name=None, is_active=None, snapshot=None,
                 mqttuserid=None, mqttpassword=None, mqttport=None,
                 mqttretainmsg=None, mqttdonotcleansession=None,
                 mqttkeepaliveinterval=None, mqttmsgmaxdeliveryattempts=None,
                 mqttmsgdelaydeliveryattempts=None, mqttmsgwaitbeforeretry=None,
                 mqttmaxinflightmsg=None, collapse=None, rmretdel=None,
                 dateformat=None, protofile=None, protomsg=None, mqttssl=None,
                 mqttsslcafile=None, mqttsslcapath=None, mqttsslcertfile=None,
                 mqttsslkeyfile=None, mqttsslpassword=None, csvincludeschema=None,
                 configfilesection=None, mqttpasswordencrypted=None,
                 addcsvopcode=None, addcsvflags=None, csvmsgperevent=None,
                 csvmsgpereventblock=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'mqtt', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['mqtthost', 'mqttclientid',
                                                   'mqtttopic', 'mqttqos',
                                                   'mqttmsgtype'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4],
                   name=name, is_active=is_active, **properties)


class MQTTPublisher(Connector):
    '''
    Subscribe to MQ Telemetry Transport (MQTT)

    Parameters
    ----------
    mqtthost : string
        Specifies the MQTT server host name
    mqttclientid : string
        Specifies the string to use as the MQTT Client ID. If NULL, a random
        client ID is generated. If empty, mqttdonotcleansession must be
        false. Must be unique among all clients connected to the MQTT server.
    mqtttopic : string
        Specifies the string to use as an MQTT subscription topic pattern
    mqttqos : string
        Specifies the requested Quality of Service. Values can be 0, 1 or 2. 
    mqttmsgtype : string
        Specifies binary, CSV, JSON, or opaquestring
    mqttuserid : string, optional
        Specifies the user name required to authenticate the connector’s
        session with the MQTT server.
    mqttpassword : string, optional
        Specifies the password associated with mqttuserid
    mqttport : int, optional
        Specifies the MQTT server port. Default is 1883.
    mqttacceptretainedmsg : boolean, optional
        Sets to true to accept to receive retained message. Default is false.
    mqttcleansession : boolean, optional
        Set to true to instruct the MQTT Server to clean all messages and
        subscriptions on disconnect, false to instruct it to keep them.
        Default is true.
    mqttkeepaliveinterval : int, optional
        Specifies the number of seconds after which the broker should send
        a PING message to the client if no other messages have been
        exchanged in that time. Default is 10.
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert
    transactional : string, optional
        When mqttmsgtype=CSV, sets the event block type to transactional.
        The default value is normal.
    blocksize : int, optional
        When mqttmsgtype=CSV, specifies the number of events to include in
        a published event block. The default value is 1.
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be parsed,
        the event is dropped, an error is logged, and publishing continues.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input CSV events.
        The default delimiter is the , character.
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that is
        automatically generated by the Source window.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted as
        an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition used to convert event blocks to protobuf messages.
        When you specify this parameter, you must also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the .proto
        file that you specified with the protofile parameter. Event blocks
        are converted into this message.
    mqttssl : boolean, optional
        Specifies to use SSL/TLS to connect to the MQTT broker. Default is
        false. In order to use SSL/TLS, the ESP encryption overlay must
        be installed.
    mqttsslcafile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM
        encoded trusted CA certificate files. Either mqttsslcafile or
        mqttsslcapath must be specified.
    mqttsslcapath : string, optional
        If mqttssl=true, specifies the path to a directory containing the
        PEM encoded trusted CA certificate files. See mosquitto.conf for
        more details about configuring this directory. Either mqttsslcafile
        or mqttsslcapath must be specified.
    mqttsslcertfile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM
        encoded certificate file for this client. Both mqttsslcertfile and
        mqttsslkeyfile must be provided if one of them is.
    mqttsslkeyfile : string, optional
        If mqttssl=true, specifies the path to a file containing the PEM
        encoded private key for this client. Both mqttsslcertfile and
        mqttsslkeyfile must be provided if one of them is.
    mqttsslpassword : string, optional
        If mqttssl=true, and if key file is encrypted, specifies the
        password for decryption.
    addcsvopcode : boolean, optional
        Prepends an opcode and comma to input CSV events. The opcode is Insert
        unless publishwithupsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events (with a comma).
        Valid values are "normal" and "partialupdate".
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].
    mqttpasswordencrypted : boolean, optional
        Specifies that mqttpassword is encrypted
    maxevents : int, optional
         Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`MQTTPublisher`

    '''
    connector_key = dict(cls='mqtt', type='publish')

    property_defs = dict(
        mqtthost=prop('mqtthost', dtype='string', required=True),
        mqttclientid=prop('mqttclientid', dtype='string', required=True),
        mqtttopic=prop('mqtttopic', dtype='string', required=True),
        mqttqos=prop('mqttqos', dtype='string', required=True),
        mqttmsgtype=prop('mqttmsgtype', dtype='string', required=True),
        mqttuserid=prop('mqttuserid', dtype='string'),
        mqttpassword=prop('mqttpassword', dtype='string'),
        mqttport=prop('mqttport', dtype='int'),
        mqttacceptretainedmsg=prop('mqttacceptretainedmsg', dtype='boolean'),
        mqttcleansession=prop('mqttcleansession', dtype='boolean'),
        mqttkeepaliveinterval=prop('mqttkeepaliveinterval', dtype='int'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        mqttssl=prop('mqttssl', dtype='boolean'),
        mqttsslcafile=prop('mqttsslcafile', dtype='string'),
        mqttsslcapath=prop('mqttsslcapath', dtype='string'),
        mqttsslcertfile=prop('mqttsslcertfile', dtype='string'),
        mqttsslkeyfile=prop('mqttsslkeyfile', dtype='string'),
        mqttsslpassword=prop('mqttsslpassword', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='string'),
        mqttpasswordencrypted=prop('mqttpasswordencrypted', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, mqtthost=None, mqttclientid=None, mqtttopic=None,
                 mqttqos=None, mqttmsgtype=None, name=None, is_active=None,
                 mqttuserid=None, mqttpassword=None, mqttport=None,
                 mqttretainmsg=None, mqttcleansession=None,
                 mqttkeepaliveinterval=None, publishwithupsert=None,
                 transactional=None, blocksize=None, ignorecsvparseerrors=None,
                 csvfielddelimiter=None, noautogenfield=None,
                 dateformat=None, protofile=None, protomsg=None,
                 mqttssl=None, mqttsslcafile=None, mqttsslcapath=None,
                 mqttsslcertfile=None, mqttsslkeyfile=None,
                 mqttsslpassword=None, configfilesection=None,
                 addcsvopcode=None, addcsvflags=None,
                 mqttpasswordencrypted=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'mqtt', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['mqtthost', 'mqttclientid',
                                                   'mqtttopic', 'mqttqos',
                                                   'mqttmsgtype'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4],
                   name=name, is_active=is_active, **properties)
